import gc
import logging
import os
import pickle
import tempfile

import numpy
from orangecanvas.scheme.readwrite import literal_dumps
from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt

from orangecontrib.tomwer.widgets.reconstruction.DarkRefAndCopyOW import (
    DarkRefAndCopyOW,
)
from tomwer.core.utils.scanutils import MockNXtomo

logging.disable(logging.INFO)


class TestDarkRefWidget(TestCaseQt):
    """class testing the DarkRefWidget"""

    def setUp(self):
        self._tmp_path = tempfile.mkdtemp()
        scan_folder_with_raw = os.path.join(self._tmp_path, "test_dir_1")
        scan_folder_without_raw = os.path.join(self._tmp_path, "test_dir_2")
        for my_dir in (scan_folder_with_raw, scan_folder_without_raw):
            os.makedirs(my_dir)

        self._scan_with_raw = MockNXtomo(
            scan_path=scan_folder_with_raw,
            create_ini_dark=True,
            create_ini_flat=True,
            create_final_flat=False,
            n_proj=10,
            n_ini_proj=10,
            dim=12,
        ).scan
        self._scan_without_raw = MockNXtomo(
            scan_path=scan_folder_without_raw,
            create_ini_dark=False,
            create_ini_flat=False,
            create_final_flat=False,
            n_proj=10,
            n_ini_proj=10,
            dim=12,
        ).scan

        self.widget = DarkRefAndCopyOW()
        return super().setUp()

    def tearDown(self):
        # shutil.rmtree(self._tmp_path)
        self.widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.widget.close()
        gc.collect()

    def test_serializing(self):
        self.widget._updateSettingsVals()
        pickle.dumps(self.widget._ewoks_default_inputs)

    def test_literal_dumps(self):
        self.widget._updateSettingsVals()
        literal_dumps(self.widget._ewoks_default_inputs)

    def test_copy(self):
        self.widget.setCopyActive(True)
        while self.qapp.hasPendingEvents():
            self.qapp.processEvents()
        assert len(os.listdir(self.widget._processing_stack._save_dir)) == 0

        self.widget.process(self._scan_with_raw)
        self.widget._processing_stack.wait_computation_finished()
        while self.qapp.hasPendingEvents():
            self.qapp.processEvents()
        assert self._scan_with_raw.load_reduced_darks() not in (None, {})
        assert self._scan_with_raw.load_reduced_flats() not in (None, {})
        assert len(os.listdir(self.widget._processing_stack._save_dir)) == 1
        assert len(os.listdir(self.widget._processing_stack._save_dir)) == 1

        self.widget.setCopyActive(False)
        while self.qapp.hasPendingEvents():
            self.qapp.processEvents()
        self.widget.process(self._scan_without_raw)
        self.widget._processing_stack.wait_computation_finished()
        while self.qapp.hasPendingEvents():
            self.qapp.processEvents()
        assert self._scan_without_raw.load_reduced_darks() in (None, {})
        assert self._scan_without_raw.load_reduced_flats() in (None, {})

        self.widget.setCopyActive(True)
        while self.qapp.hasPendingEvents():
            self.qapp.processEvents()
        self.widget.process(self._scan_without_raw)
        self.widget._processing_stack.wait_computation_finished()
        while self.qapp.hasPendingEvents():
            self.qapp.processEvents()
        assert self._scan_without_raw.load_reduced_darks() not in (None, {})
        assert self._scan_without_raw.load_reduced_flats() not in (None, {})

    def test_set_darks_and_flats(self):
        """
        Test settings darks and flats.
        """
        self.widget.setCopyActive(True)

        assert len(os.listdir(self.widget._processing_stack._save_dir)) == 0
        self.widget.setReducedFlats(
            {
                "0.0r": numpy.linspace(12, 12 + 12 * 12, 12 * 12).reshape(12, 12),
                "1.0r": numpy.linspace(24, 24 + 12 * 12, 12 * 12).reshape(12, 12),
            }
        )
        self.widget.setReducedDarks(
            {
                "0.0r": numpy.linspace(0, 12 * 12, 12 * 12).reshape(12, 12),
            }
        )
        self.widget.process(self._scan_without_raw)
        self.widget._processing_stack.wait_computation_finished()
        while self.qapp.hasPendingEvents():
            self.qapp.processEvents()

        loaded_reduced_darks = self._scan_without_raw.load_reduced_darks()
        assert tuple(loaded_reduced_darks.keys()) == (0,)
        numpy.testing.assert_allclose(
            loaded_reduced_darks[0],
            numpy.linspace(0, 12 * 12, 12 * 12).reshape(12, 12),
        )

        loaded_reduced_flats = self._scan_without_raw.load_reduced_flats()
        assert tuple(loaded_reduced_flats.keys()) == (0, 9)
        numpy.testing.assert_allclose(
            loaded_reduced_flats[9],
            numpy.linspace(24, 24 + 12 * 12, 12 * 12).reshape(12, 12),
        )



from __future__ import annotations

import numpy
from silx.gui import qt
from silx.gui.utils import concurrent

from silx.gui.widgets.WaitingOverlay import WaitingOverlay
import logging

from tomoscan.volumebase import SliceTuple

from tomwer.gui import icons as tomwer_icons
from tomwer.gui.visualization.TomoPlot2D import TomoPlot2D as _TomoPlot2D

from tomwer.core.volume.volumebase import TomwerVolumeBase

from ._HorizontalSliderWithBrowser import HorizontalSliderWithBrowser
from .CoordinateSystemOverlay import CoordinateSystemOverlay
from ._ProcessingOverlay import ProcessingOverlay
from ._SliceLoader import SliceLoader

_logger = logging.getLogger(__name__)


class TomoPlot2DSlices(qt.QWidget):
    """
    A composition of a TomoPlot2D, a QLabel (with a colored frame) and a set of radio buttons for the user to select the index to be displayed

    :param title: plot title.
    :param color: color associated to the plot.
    :param image: image to associate for the overlay coordinate system.
    :param axis: Axis associated to the plot (determine how to read from the raw volume).
    """

    displayFullScreen = qt.Signal()

    def __init__(self, title: str, color: str, image: str, axis: int, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if axis not in (0, 1, 2):
            raise ValueError(f"'axis' should be in (0, 1, 2). Got {type(axis)}")
        self._axis = axis
        self.__sliceLoader: SliceLoader | None = None
        # thread used to load a slice when not part of the 'slices' and when the volume data is not in memory.
        # avoid any deadlock or GUI lag.

        self.setLayout(qt.QVBoxLayout())
        self.layout().setContentsMargins(2, 2, 2, 2)
        self.layout().setSpacing(2)

        self._slices: dict[SliceTuple, numpy.ndarray] = {}
        # sub set of slice easily accessible
        self._volume: TomwerVolumeBase | None = None
        # optional volume from which we can read slices.
        # The volume is expected to be loaded or reading is expected to be done on the 'fast' axis.

        self._plot = _TomoPlot2D(parent=self)
        self._plot.getColorBarWidget().hide()
        self._plot.getOutputToolBar().hide()
        self._plot.getColormapAction().setVisible(False)
        self._plot.getColorBarAction().setVisible(False)
        self._plot.getMaskAction().setVisible(False)
        self._plot.getInteractiveModeToolBar().hide()
        self.layout().addWidget(self._plot)

        self._coordinateSystemOverlay = CoordinateSystemOverlay(
            parent=self._plot.getWidgetHandle(),
            img=image,
            img_size=qt.QSize(50, 50),
        )
        self._coordinateSystemOverlay.setAlignment(qt.Qt.AlignRight | qt.Qt.AlignBottom)
        self._coordinateSystemOverlay.setAlignmentOffsets((-10, -10))

        self._title = qt.QLabel(title, parent=self)
        self._title.setAlignment(qt.Qt.AlignCenter)
        self.layout().addWidget(self._title)

        self._sliceIndexSelection = HorizontalSliderWithBrowser()
        self.layout().addWidget(self._sliceIndexSelection)

        # set the color to the title frame (simplest)
        self._title.setStyleSheet(
            f"border: 2px solid {color}",
        )

        # overlay to signal some loading is on-going.
        self._messageOverlayWidget = WaitingOverlay(self._plot)
        self._messageOverlayWidget.setStyleSheet(
            "QPushButton { background-color: rgba(150, 150, 150, 40); border: 0px; border-radius: 10px; }"
        )
        self._messageOverlayWidget.setIconSize(qt.QSize(30, 30))

        # Overlay to notify during fast axis reading
        self._fastReadingIndicator = ProcessingOverlay(
            self._plot, img_size=qt.QSize(5, 6)
        )
        self._fastReadingIndicator.setAlignmentOffsets((-10, 10))
        self._fastReadingIndicator.setVisible(False)

        # create the new toolbar
        # full screen action
        self._extraToolbar = qt.QToolBar(self)
        self._plot.addToolBar(self._extraToolbar)
        fullScreenIcon = tomwer_icons.getQIcon("full_screen")
        self._fullScreenAction = qt.QAction(fullScreenIcon, "pop up full screen")
        self._extraToolbar.addAction(self._fullScreenAction)

        self._toolbars = (
            self._plot.toolBar(),
            self._plot.getProfileToolbar(),
            self._extraToolbar,
        )

        # set up
        self.setCoordinateSystemLabelVisible(False)
        self.setToolBarsVisible(False)
        self.setLoading(False)

        # connect signal / slot
        self._fullScreenAction.triggered.connect(self.displayFullScreen)
        self._sliceIndexSelection.valueChanged.connect(self._updatePlot)

    def _indicateFastReading(self, value: bool):
        """
        At the moment the fast-reading axis can become load in the case of hdf5. At the moment we simply notify it to the user.
        """
        self._fastReadingIndicator.setVisible(value)

    def setLoadingMessage(self, message: str):
        self._messageOverlayWidget.setText(message)

    def setLoading(self, loading: bool) -> None:
        self._messageOverlayWidget.setVisible(loading)

    def setCoordinateSystemVisible(self, visible):
        self._coordinateSystemOverlay.setVisible(visible)

    def setCoordinateSystemLabelVisible(self, visible: bool):
        self._title.setVisible(visible)

    def setSliceSelectionVisible(self, visible: bool):
        self._sliceIndexSelection.setVisible(visible)

    def setToolBarsVisible(self, visible: bool):
        for toolbar in self._toolbars:
            toolbar.setVisible(visible)

    def getTitle(self) -> str:
        return self._title

    def setVolume(self, volume: TomwerVolumeBase | None):
        """
        Set the volume. So user can browse through it in addition to the slices.
        """
        if not isinstance(volume, (TomwerVolumeBase, type(None))):
            raise TypeError(
                f"'volume' is expected to be instance of {TomwerVolumeBase} or {None}. Got {type(volume)}."
            )
        self._volume = volume
        if volume is not None:
            # avoid updating the range (and the slice index) when possible
            range = 0, self._volume.get_volume_shape()[self._axis] - 1
            if range != self._sliceIndexSelection.getRange():
                self._sliceIndexSelection.setRange(*range)
        self._sliceIndexSelection.setSliceBrowsingEnabled(self._volume is not None)

    @staticmethod
    def _determinePreferredSliceIndex(slice_indices: tuple[int]) -> int:
        """Return the index to display if None provided during construction"""
        if len(slice_indices) == 0:
            return -1
        return slice_indices[len(slice_indices) // 2]

    def setSlices(
        self, slices: dict[SliceTuple, numpy.ndarray], update_index: bool = True
    ):
        self._slices = {
            slice_tuple.index: value for slice_tuple, value in slices.items()
        }
        # update to the new one
        slice_indices = tuple(self._slices.keys())
        self._sliceIndexSelection.setSliceIndices(slices=slice_indices)
        if update_index:
            self._sliceIndexSelection.setValue(
                self._determinePreferredSliceIndex(slice_indices=slice_indices)
            )

        self._updatePlot(slice_index=self._sliceIndexSelection.value())

    def setSliceRange(self, first, last):
        self._sliceIndexSelection.setRange(first, last)

    def getSliceRange(self):
        return self._sliceIndexSelection.getRange()

    def _updatePlot(self, slice_index: int | None):
        if slice_index is None:
            self._plot.clear()
            return

        image = self._slices.get(slice_index, None)
        loading_delegated = False
        if image is None and self._volume is not None:
            if self._volume.data is None:
                loading_delegated = self.__delegateSliceLoading(slice_index=slice_index)
            else:
                image = self._volume.get_slice(
                    index=slice_index,
                    axis=self._axis,
                )

        if loading_delegated:
            return

        if image is None:
            _logger.error(f"Fail to load slice index {slice_index}")
            return

        self._plot.addImage(image, resetzoom=False)

    def __delegateSliceLoading(self, slice_index: int) -> bool:
        """Delegate slice loading to a dedicated thread"""
        if self.__sliceLoader is not None:
            self.__sliceLoader.finished.disconnect(self.__delegatedSliceLoadingFinished)
            # TODO: fix me: avoid some segfault having several thread reading and being destroyed without owner.
            # wait for the previous loading to be done.
            # This avoid having several thread and ensure a 'clean' loading of the data.
            # Remember that we are on the 'fast' reading axis.
            # self.__sliceLoader.wait()
        self._indicateFastReading(True)
        self.__sliceLoader = SliceLoader(
            parent=self,
            volume=self._volume,
            slice_index=slice_index,
            axis=self._axis,
        )
        self.__sliceLoader.finished.connect(self.__delegatedSliceLoadingFinished)
        self.__sliceLoader.start()
        return True

    def __delegatedSliceLoadingFinished(self):
        if self.sender() != self.__sliceLoader:
            # slice to be loaded has been changed
            return

        self._indicateFastReading(False)
        self.__sliceLoader.finished.disconnect(self.__delegatedSliceLoadingFinished)
        data = self.__sliceLoader.data
        slice_index = self.__sliceLoader.slice_index

        if data is not None:
            if slice_index == self.getCurrentSliceIndex():
                concurrent.submitToQtMainThread(
                    self._plot.addImage,
                    data,
                    resetzoom=False,
                )
            else:
                # can happen if not being responsive enough (volume being load, user browsing the fast axis and the loading of slices along the fast axis not being so fast...)
                pass
        else:
            _logger.warning(f"Failed to load slice {self.__sliceLoader.slice_index}")
        self.__sliceLoader = None

    def getPlot(self):
        return self._plot

    # expose API
    def clear(self):
        self._plot.clear()

    def setVoxelSize(self, size_m: float | None):
        self._plot.setVoxelSize(size=size_m)

    def setCurrentIndex(self, index: int):
        self._sliceIndexSelection.setValue(index)

    def getCurrentSliceIndex(self) -> int:
        return self._sliceIndexSelection.value()



from tomwer.tests.conftest import qtapp  # noqa F401
from tomwer.gui.dataportal.gallery import GalleryWidget


def test_GalleryWidget(
    qtapp,  # noqa F811
):
    widget = GalleryWidget()
    assert widget.getConfiguration() == {
        "output_format": "png",
        "overwrite": True,
        "binning": "16x16",
    }

    new_config = {
        "output_format": "jpg",
        "overwrite": False,
        "binning": "4x4",
    }
    widget.setConfiguration(new_config)
    assert widget.getConfiguration() == new_config



from silx.gui import qt


class NXtomoConcatenateWidget(qt.QWidget):
    """
    Widget to define setting of a concatenation of NXtomo.
    User need to provide entry name (data_path), output file path (a default option is possible) and if they want to overwrite any existing file / entry if necessary
    """

    sigConfigChanged = qt.Signal()
    """Emitted on configuration modifications"""

    class OutputWidget(qt.QWidget):
        def __init__(self, parent=None, *args, **kwargs) -> None:
            super().__init__(parent, *args, **kwargs)
            self.setLayout(qt.QFormLayout())
            self.layout().setContentsMargins(0, 0, 0, 0)
            self._outputFilePathQLE = qt.QLineEdit("{common_path}/concatenate.nx", self)
            self._outputFilePathQLE.setToolTip(
                """location of the concatenation. Possible keywords are: \n
                - 'common_path': common path of all the scans to concatenate
                """
            )
            self.layout().addRow("output file path", self._outputFilePathQLE)
            self._outputDataPathQLE = qt.QLineEdit("entry0000", self)
            self.layout().addRow("output entry", self._outputDataPathQLE)

    def __init__(self, parent=None, *args, **kwargs):
        super().__init__(parent=parent, *args, **kwargs)
        self.setLayout(qt.QVBoxLayout())
        # output widget
        self._outputWidget = self.OutputWidget(parent=self)
        self.layout().addWidget(self._outputWidget)
        # overwrite option
        self._overwriteCB = qt.QCheckBox("overwrite", parent)
        self.layout().addWidget(self._overwriteCB)
        # spacer
        spacer = qt.QWidget(parent=self)
        spacer.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Expanding)
        self.layout().addWidget(spacer)

        # conenct signal / slot
        self._outputWidget._outputFilePathQLE.textEdited.connect(self._changed)
        self._outputWidget._outputDataPathQLE.textEdited.connect(self._changed)
        self._overwriteCB.toggled.connect(self._changed)

    def _changed(self, *args, **kwargs):
        self.sigConfigChanged.emit()

    def getConfiguration(self) -> dict:
        return {
            "overwrite": self._overwriteCB.isChecked(),
            "output_file": self.getOutputFilePath(),
            "output_entry": self.getOutputEntry(),
        }

    def setConfiguration(self, config: dict) -> None:
        overwrite = config.get("overwrite", None)
        if overwrite is not None:
            self._overwriteCB.setChecked(overwrite in (True, "True", "1", "true"))

        output_file = config.get("output_file", None)
        if output_file is not None:
            self._outputWidget._outputFilePathQLE.setText(output_file)

        output_entry = config.get("output_entry", None)
        if output_entry is not None:
            self._outputWidget._outputDataPathQLE.setText(output_entry)

    # expose API
    def getOutputFilePath(self) -> str:
        return self._outputWidget._outputFilePathQLE.text()

    def getOutputEntry(self) -> str:
        return self._outputWidget._outputDataPathQLE.text()



from __future__ import annotations

from tomwer.core.scan.scanbase import TomwerScanBase
from collections import OrderedDict

_CACHE: {str, tuple[float | None, dict | None]} = OrderedDict()
"""
In the case we want to reprocess a scan we (might) need to know the cor and the nabu reconstruction parameters (for nabu slice and volume reconstructions)
But we are not keeping all those object during the full 'orange canvas' lifecycle, it would be too heavy.
As a consequence we cache those (small) values (and only those) of the last n scans (n==30) in case users want to reprocess some of those.
n value is an arbitrary value.
"""
_CACHE_SIZE = 30


def save_reconstruction_parameters_to_cache(scan: TomwerScanBase):
    # pop if exists. Will move this item as 'last-in'
    _CACHE.pop(scan.get_identifier().to_str(), None)
    _CACHE[scan.get_identifier().to_str()] = (
        None if scan.axis_params is None else scan.axis_params.relative_cor_value,
        None if scan.nabu_recons_params is None else scan.nabu_recons_params,
    )
    if len(_CACHE) > _CACHE_SIZE:
        _CACHE.popitem(last=False)


def load_reconstruction_parameters_from_cache(scan: TomwerScanBase):
    cor, nabu_recons_params = _CACHE.get(scan.get_identifier().to_str(), (None, None))
    if cor is not None and scan.axis_params is not None:
        scan.axis_params.set_relative_value(cor)
    if nabu_recons_params is not None:
        scan.nabu_recons_params = nabu_recons_params


def clear_reconstruction_parameters_cache():
    _CACHE.clear()



import logging

from enum import Enum as _Enum
from tomwer.core.process.reconstruction.axis.side import Side

_logger = logging.getLogger(__name__)


class AxisModeMetadata:
    """
    Util class to store metadata regarding processing of a mode.

    Ease processing and display
    Maybe this function should be part of nabu ?

    :param lockable: True if this value can be lock and process automatically when a new scan is received
    :param tooltip: short description of the mode (to be used by the GUI)
    :param computing_constrains: some constrain like need to have a 0-360 acquisition (so no Half acquisition)
    :param allows_padding: does the algorithm applies padding (and so can the user provide some mode)
    :param valid_inputs: compatible input type (pair of radios or sinogram...)
    :param valid_sides: can a side be provided
    :param allows_estimated_cor_as_numerical_value: can a numerical value (first guess) be provided to the algorithm
    """

    def __init__(
        self,
        lockable,
        tooltip="",
        computing_constrains=(),
        allows_padding=False,
        valid_inputs=(),
        valid_sides=(),
        allows_estimated_cor_as_numerical_value: bool = True,
    ) -> None:
        self._lockable = lockable
        self._tooltip = tooltip
        self._computing_constrains = computing_constrains
        self._allows_padding = allows_padding
        self._valid_inputs = valid_inputs
        self._valid_sides = valid_sides
        self._allows_estimated_cor_as_numerical_value = (
            allows_estimated_cor_as_numerical_value
        )

    @property
    def is_lockable(self) -> bool:
        return self._lockable

    @property
    def tooltip(self) -> str:
        return self._tooltip

    @property
    def computing_constrains(self) -> tuple:
        return self._computing_constrains

    @property
    def allows_padding(self) -> bool:
        return self._allows_padding

    @property
    def valid_inputs(self) -> tuple:
        return self._valid_inputs

    @property
    def valid_sides(self) -> tuple:
        return self._valid_sides

    @property
    def allows_estimated_cor_as_numerical_value(self) -> bool:
        return self._allows_estimated_cor_as_numerical_value


class _InputType(_Enum):
    SINOGRAM = "sinogram"
    RADIOS_X2 = "2 radios"
    COMPOSITE = "composite"


class _Constrain(_Enum):
    FULL_TURN = "full turn"


class AxisMode(_Enum):
    centered = "centered"
    global_ = "global"
    manual = "manual"
    growing_window_sinogram = "sino-growing-window"
    growing_window_radios = "growing-window"
    sliding_window_sinogram = "sino-sliding-window"
    sliding_window_radios = "sliding-window"
    sino_coarse_to_fine = "sino-coarse-to-fine"
    composite_coarse_to_fine = "composite-coarse-to-fine"
    fourier_angles = "fourier-angles"
    octave_accurate_radios = "octave-accurate"
    read = "read 'x_rotation_axis_pixel_position'"
    # alias to composite_coarse_to_fine with near mode
    near = "near"

    @classmethod
    def from_value(cls, value):
        # ensure backward compatibility with workflow defined before COR method on sinograms
        if value in ("global_", "global"):
            value = AxisMode.global_
        if value == "radio-growing-window":
            _logger.warning(
                f"Axis mode requested is '{value}'. To insure backward compatibility replace it by '{AxisMode.growing_window_radios.value}'"
            )
            value = AxisMode.growing_window_radios
        elif value == "radio-sliding-window":
            _logger.warning(
                f"Axis mode requested is '{value}'. To insure backward compatibility replace it by '{AxisMode.sliding_window_radios.value}'"
            )
            value = AxisMode.sliding_window_radios
        elif value in ("radios-octave-accurate", "accurate"):
            _logger.warning(
                f"Axis mode requested is '{value}'. To insure backward compatibility replace it by '{AxisMode.octave_accurate_radios.value}'"
            )
            value = AxisMode.octave_accurate_radios
        elif value in ("read", "read from estimated cor"):
            value = AxisMode.read

        return AxisMode(value)

    def requires_radio_indices(self) -> bool:
        return self in (
            AxisMode.growing_window_radios,
            AxisMode.sliding_window_radios,
            AxisMode.octave_accurate_radios,
        )

    def requires_sinogram_index(self) -> bool:
        return self in (
            AxisMode.growing_window_sinogram,
            AxisMode.sliding_window_sinogram,
            AxisMode.fourier_angles,
            AxisMode.sino_coarse_to_fine,
        )


AXIS_MODE_METADATAS = {
    # manual
    AxisMode.manual: AxisModeMetadata(
        lockable=False,
        tooltip="Enter or find manually the COR value",
        computing_constrains=(),
        allows_padding=False,
        valid_inputs=(_InputType.RADIOS_X2,),
        allows_estimated_cor_as_numerical_value=False,
    ),
    # read
    AxisMode.read: AxisModeMetadata(
        lockable=True,
        tooltip="Read COR value from nexus file ({entry}/instrument/detector/x_rotation_axis_pixel_position dataset). Will work only for NXtomo / hdf5 datasets",
        computing_constrains=(),
        allows_padding=False,
        valid_inputs=None,
        allows_estimated_cor_as_numerical_value=False,
    ),
    # radio algorithm
    AxisMode.centered: AxisModeMetadata(
        lockable=True,
        tooltip="Dedicated to fullfield. Previously named 'accurate'",
        computing_constrains=(),
        allows_padding=True,
        valid_inputs=(_InputType.RADIOS_X2,),
        valid_sides=(Side.CENTER,),
    ),
    AxisMode.global_: AxisModeMetadata(
        lockable=True,
        tooltip="Algorithm which can work for both half acquisition and standard ('full field') acquisition",
        computing_constrains=(),
        allows_padding=True,
        valid_inputs=(_InputType.RADIOS_X2,),
        allows_estimated_cor_as_numerical_value=False,
    ),
    AxisMode.growing_window_radios: AxisModeMetadata(
        lockable=True,
        tooltip="A auto-Cor method",
        computing_constrains=(),
        allows_padding=True,
        valid_inputs=(_InputType.RADIOS_X2,),
        valid_sides=(Side.RIGHT, Side.LEFT, Side.CENTER, Side.ALL),
        allows_estimated_cor_as_numerical_value=False,
    ),
    AxisMode.sliding_window_radios: AxisModeMetadata(
        lockable=True,
        tooltip="A method for estimating semi-automatically the CoR position. You have to provide a hint on where the CoR is (left, center, right).",
        computing_constrains=(),
        allows_padding=True,
        valid_inputs=(_InputType.RADIOS_X2,),
        valid_sides=(Side.RIGHT, Side.LEFT, Side.CENTER),
    ),
    AxisMode.octave_accurate_radios: AxisModeMetadata(
        lockable=True,
        tooltip="Same method as the 'accurate' octave code",
        computing_constrains=(_Constrain.FULL_TURN,),
        allows_padding=True,
        valid_inputs=(_InputType.RADIOS_X2,),
        valid_sides=(Side.CENTER,),
        allows_estimated_cor_as_numerical_value=False,
    ),
    # sinogram algorithm
    AxisMode.growing_window_sinogram: AxisModeMetadata(
        lockable=True,
        tooltip="A auto-Cor method",
        computing_constrains=(),
        allows_padding=True,
        valid_inputs=(_InputType.SINOGRAM,),
        valid_sides=(Side.RIGHT, Side.LEFT, Side.CENTER, Side.ALL),
        allows_estimated_cor_as_numerical_value=False,
    ),
    AxisMode.sliding_window_sinogram: AxisModeMetadata(
        lockable=True,
        tooltip="A method for estimating semi-automatically the CoR position. You have to provide a hint on where the CoR is (left, center, right).",
        computing_constrains=(),
        allows_padding=True,
        valid_inputs=(_InputType.SINOGRAM,),
        valid_sides=(
            Side.RIGHT,
            Side.LEFT,
            Side.CENTER,
        ),
    ),
    AxisMode.sino_coarse_to_fine: AxisModeMetadata(
        lockable=True,
        tooltip="Estimate CoR from sinogram. Only works for 360 degrees scans.",
        computing_constrains=(_Constrain.FULL_TURN,),
        allows_padding=True,
        valid_inputs=(_InputType.SINOGRAM,),
        valid_sides=(
            Side.RIGHT,
            Side.LEFT,
        ),
        allows_estimated_cor_as_numerical_value=False,
    ),
    AxisMode.fourier_angles: AxisModeMetadata(
        lockable=True,
        tooltip="",
        computing_constrains=(_Constrain.FULL_TURN,),
        allows_padding=True,
        valid_inputs=(_InputType.SINOGRAM,),
        valid_sides=(
            Side.RIGHT,
            Side.LEFT,
            Side.CENTER,
        ),
    ),
    # coarse-to-fine algorithm
    AxisMode.composite_coarse_to_fine: AxisModeMetadata(
        lockable=True,
        tooltip="A auto-Cor method",
        computing_constrains=(_Constrain.FULL_TURN,),
        allows_padding=True,
        valid_inputs=(_InputType.COMPOSITE,),
        valid_sides=(
            Side.RIGHT,
            Side.LEFT,
            Side.CENTER,
        ),
    ),
    AxisMode.near: AxisModeMetadata(
        lockable=True,
        tooltip="Alias to composite_coarse_to_fine",
        computing_constrains=(_Constrain.FULL_TURN,),
        allows_padding=True,
        valid_inputs=(_InputType.COMPOSITE,),
        valid_sides=(
            Side.RIGHT,
            Side.LEFT,
            Side.CENTER,
        ),
    ),
}



# coding: utf-8
from __future__ import annotations


import os
import shutil
import tempfile
import unittest

from tomwer.core.process.reconstruction.sadeltabeta.sadeltabeta import (
    SADeltaBetaParams,
    SADeltaBetaTask,
)
from tomwer.core.utils.scanutils import MockNXtomo


class TestSADeltaBetaProcess(unittest.TestCase):
    """Test the SAAxisProcess class"""

    def setUp(self) -> None:
        super().setUp()
        self.tempdir = tempfile.mkdtemp()
        dim = 10
        mock = MockNXtomo(
            scan_path=self.tempdir, n_proj=10, n_ini_proj=10, scan_range=180, dim=dim
        )
        self.scan = mock.scan

    def tearDown(self) -> None:
        shutil.rmtree(self.tempdir)
        super().tearDown()

    def test(self):
        process = SADeltaBetaTask(
            inputs={
                "data": self.scan,
                "sa_delta_beta_params": SADeltaBetaParams().to_dict(),
                "serialize_output_data": False,
            }
        )

        default_sadelta_beta_params = SADeltaBetaParams()
        default_sadelta_beta_params.output_dir = os.path.join(
            self.tempdir, "output_dir"
        )
        default_sadelta_beta_params.dry_run = True

        process.run()



# coding: utf-8

from __future__ import annotations

import copy
import logging
from contextlib import AbstractContextManager

from orangewidget import gui
from orangewidget.settings import Setting
from orangewidget.widget import Input, Output
from silx.gui import qt

from tomwer.core.process.reconstruction.params_cache import (
    load_reconstruction_parameters_from_cache,
    save_reconstruction_parameters_to_cache,
)
import tomwer.core.process.reconstruction.nabu.nabuslices
from tomwer.core.cluster import SlurmClusterConfiguration
from tomwer.core.futureobject import FutureTomwerObject
from tomwer.core.scan.scanbase import TomwerScanBase, _TomwerBaseDock
from tomwer.core.volume.volumefactory import VolumeFactory

from tomwer.gui.reconstruction.nabu.slices import NabuWindow
from tomwer.synctools.stacks.reconstruction.nabu import NabuSliceProcessStack

from ...orange.managedprocess import SuperviseOW
from ..utils import WidgetLongProcessing

_logger = logging.getLogger(__name__)


class NabuOW(WidgetLongProcessing, SuperviseOW):
    """
    A simple widget managing the copy of an incoming folder to an other one

    :param parent: the parent widget
    """

    # note of this widget should be the one registered on the documentation
    name = "nabu slice reconstruction"
    id = "orange.widgets.tomwer.reconstruction.NabuOW.NabuOW"
    description = "This widget will call nabu for running a reconstruction "
    icon = "icons/nabu_2d.svg"
    priority = 12
    keywords = ["tomography", "nabu", "reconstruction", "FBP", "filter"]

    want_main_area = True
    resizing_enabled = True

    _ewoks_default_inputs = Setting({"data": None, "nabu_params": None})

    ewokstaskclass = tomwer.core.process.reconstruction.nabu.nabuslices.NabuSlicesTask

    sigScanReady = qt.Signal(TomwerScanBase)
    "Signal emitted when a scan is ended"

    class Inputs:
        reprocess = Input(
            name="change recons params",
            type=_TomwerBaseDock,
            doc="reconpute slice with different parameters",
        )
        data = Input(
            name="data",
            type=TomwerScanBase,
            doc="one scan to be process",
            default=True,
            multiple=True,
        )
        cluster_in = Input(
            name="cluster_config",
            type=SlurmClusterConfiguration,
            doc="slurm cluster to be used",
            multiple=False,
        )

    class Outputs:
        data = Output(name="data", type=TomwerScanBase, doc="one scan to be process")

        future_out = Output(
            name="future_tomo_obj",
            type=FutureTomwerObject,
            doc="data with some remote processing",
        )

        slice_urls = Output(name="slice urls", type=tuple, doc="tuple of urls created")

    class DialogCM(AbstractContextManager):
        """Simple context manager to hide / show button dialogs"""

        def __init__(self, dialogButtonsBox):
            self._dialogButtonsBox = dialogButtonsBox

        def __enter__(self):
            self._dialogButtonsBox.show()

        def __exit__(self, exc_type, exc_val, exc_tb):
            self._dialogButtonsBox.hide()

    def __init__(self, parent=None):
        SuperviseOW.__init__(self, parent)
        WidgetLongProcessing.__init__(self)
        self._slurmCluster = None
        self.__exec_for_ci = False
        # processing tool
        self._processingStack = NabuSliceProcessStack(self, process_id=self.process_id)

        _layout = gui.vBox(self.mainArea, self.name).layout()
        # main widget
        self._nabuWidget = NabuWindow(parent=self)
        _layout.addWidget(self._nabuWidget)
        # add button to validate when change reconstruction parameters is
        # called
        types = qt.QDialogButtonBox.Ok
        self._buttons = qt.QDialogButtonBox(self)
        self._buttons.setStandardButtons(types)
        _layout.addWidget(self._buttons)

        # set up
        self._buttons.hide()

        # load settings
        nabu_params = self._ewoks_default_inputs.get("nabu_params", None)
        if nabu_params not in (dict(), None):
            try:
                self._nabuWidget.setConfiguration(nabu_params)
            except Exception:
                _logger.warning("fail to load reconstruction settings")

        # expose API
        self.getMode = self._nabuWidget.getMode
        self.setMode = self._nabuWidget.setMode

        # connect signal / slot
        self._processingStack.sigComputationStarted.connect(self._startProcessing)
        self._processingStack.sigComputationEnded.connect(self._endProcessing)
        self._buttons.button(qt.QDialogButtonBox.Ok).clicked.connect(self.accept)
        self._nabuWidget.sigConfigChanged.connect(self._updateSettingsVals)

    def __new__(cls, *args, **kwargs):
        # ensure backward compatibility with 'static_input'
        static_input = kwargs.get("stored_settings", {}).get("static_input", None)
        if static_input not in (None, {}):
            _logger.warning(
                "static_input has been deprecated. Will be replaced by _ewoks_default_inputs in the workflow file. Please save the workflow to apply modifications"
            )
            kwargs["stored_settings"]["_ewoks_default_inputs"] = static_input
        return super().__new__(cls, *args, **kwargs)

    @Inputs.data
    def process(self, scan, *args, **kwargs):
        assert isinstance(scan, (TomwerScanBase, type(None)))
        if scan is None:
            return
        scan_ = copy.copy(scan)
        scan_.clear_latest_reconstructions()

        _logger.info(f"add {scan} to the stack")
        # update the reconstruction mode if possible
        self._nabuWidget.setScan(scan_)
        self._processingStack.add(scan_, self.getConfiguration())

    @Inputs.reprocess
    def reprocess(self, scan):
        """Recompute nabu with different parameters"""
        # wait for user to tune the reconstruction
        if scan is None:
            return

        self._nabuWidget.setScan(scan)

        if scan.axis_params is None or scan.axis_params.relative_cor_value is None:
            # try to retrieve last computed cor value from nabu process
            if scan.axis_params is None:
                from tomwer.synctools.axis import QAxisRP

                scan.axis_params = QAxisRP()
            load_reconstruction_parameters_from_cache(scan=scan)

        self.show()
        with NabuOW.DialogCM(self._buttons):
            if self.__exec_for_ci is True:
                self._ciExec()
            else:
                if self.exec():
                    # for now The behavior for reprocessing is the sama as for processing
                    if hasattr(scan, "instance"):
                        self.process(scan.instance)
                    else:
                        self.process(scan)

    def cancel(self, scan):
        if scan is None:
            return
        if scan in self._processingStack:
            self._processingStack.remove(scan)
        if (
            self._processingStack._data_currently_computed.get_identifier().to_str()
            == scan.get_identifier().to_str()
        ):
            # stop current processing
            self._processingStack.cancel()
            # if possible process next
            if self._processingStack.can_process_next():
                self._processingStack._process_next()

    @Inputs.cluster_in
    def setCluster(self, slurm_cluster: SlurmClusterConfiguration | None):
        assert isinstance(
            slurm_cluster, (type(None), SlurmClusterConfiguration)
        ), f"Expect None of SlurmClusterConfiguration. Not {type(slurm_cluster)}"
        self._slurmCluster = slurm_cluster

    def _endProcessing(self, scan, future_tomo_obj):
        WidgetLongProcessing._endProcessing(self, scan)
        if scan is not None:
            save_reconstruction_parameters_to_cache(scan=scan)
            # send scan
            self.Outputs.data.send(scan)
            # send slice urls
            slice_urls = []
            for rec_identifier in scan.latest_reconstructions:
                slice_urls.extend(
                    VolumeFactory.from_identifier_to_vol_urls(rec_identifier)
                )
            if len(slice_urls) > 0:
                slice_urls = tuple(slice_urls)
                # provide list of reconstructed slices
                self.Outputs.slice_urls.send(slice_urls)

            self.sigScanReady.emit(scan)
        if future_tomo_obj is not None:
            # send future scan
            self.Outputs.future_out.send(future_tomo_obj)

    def setDryRun(self, dry_run):
        self._processingStack.setDryRun(dry_run)

    def _ciExec(self):
        self.activateWindow()
        self.raise_()
        self.show()

    def _replaceExec_(self):
        """used for CI, replace the exec_ call ny"""
        self.__exec_for_ci = True

    def _updateSettingsVals(self):
        self._ewoks_default_inputs = {
            "data": None,
            "nabu_params": self.getConfiguration(),
        }

    def getConfiguration(self):
        config = self._nabuWidget.getConfiguration()
        config["cluster_config"] = self._slurmCluster
        return config

    def setConfiguration(self, config):
        # ignore slurm cluster. Defined by the upper widget
        config.pop("cluster_config", None)
        self._nabuWidget.setConfiguration(config=config)

    def keyPressEvent(self, event):
        """The event has to be filtered since we have some children
        that can be edited using the 'enter' key as defining the cor manually
        (see #481)). As we are in a dialog this automatically trigger
        'accepted'. See https://forum.qt.io/topic/5080/preventing-enter-key-from-triggering-ok-in-qbuttonbox-in-particular-qlineedit-qbuttonbox/5
        """
        if event.key() not in (qt.Qt.Key_Enter, qt.Qt.Key_Return):
            super().keyPressEvent(event)



import os
import psutil
from tomwer.core import settings

MOCK_LOW_MEM = False  # if True will simulate the case the computer run into low memory

IGNORE_LOW_MEM = False


def is_low_on_memory(path=""):
    """

    :return: True if the RAM usage is more than MAX_MEM_USED (or low memory
       is simulated)
    """
    if IGNORE_LOW_MEM is True:
        return False
    if path == settings.get_dest_path():
        if settings.MOCK_LBSRAM is True:
            return MOCK_LOW_MEM
        else:
            assert os.path.isdir(path)
            return psutil.disk_usage(path).percent > settings.MAX_MEM_USED
    else:
        return (MOCK_LOW_MEM is True) or (
            os.path.exists(path)
            and (psutil.disk_usage(path).percent > settings.MAX_MEM_USED)
        )


def mock_low_memory(b=True):
    """Mock the case the computer is running into low memory"""
    global MOCK_LOW_MEM
    MOCK_LOW_MEM = b
    return psutil.virtual_memory().percent > settings.MAX_MEM_USED



# coding: utf-8
from __future__ import annotations


import os
import socket
import logging

from silx.gui import qt

from tomwer.core import settings
from tomwer.core.process.utils import LastReceivedScansDict
from tomwer.core.scan.blissscan import BlissScan
from tomwer.gui import icons as tomwericons
from tomwer.gui.control import datareacheractions as actions
from tomwer.gui.control.history import ScanHistory
from tomwer.gui.control.observations import ScanObservation
from tomwer.gui.utils.inputwidget import (
    HDF5ConfigFileSelector,
    NXTomomillOutputDirSelector,
)
from tomwer.gui.utils.host_editor import HostEditor
from tomwer.synctools.rsyncmanager import BlissSequenceRSyncWorker

from ._host import Host


_logger = logging.getLogger(__name__)


class DataListenerWidget(qt.QMainWindow):
    """
    Widget to display the bliss acquisition on going and finished
    """

    NB_STORED_LAST_FOUND = 20

    sigActivate = qt.Signal()
    """Signal emitted when the listening start"""
    sigDeactivate = qt.Signal()
    """Signal emitted when the listening end"""
    sigConfigurationChanged = qt.Signal()
    """Signal emitted when the configuration for the bliss client is updated"""
    sigAcquisitionEnded = qt.Signal(tuple)
    """Signal emitted when an acquisition is ended without errors.
    Tuple contains (master_file, entry, proposal_file)"""
    sigServerStopped = qt.Signal()
    """Signal emitted when the server is stopped by a sigkill or sigterm"""
    sigCFGFileChanged = qt.Signal(str)
    """Signal emitted when path to the nxtomomill configuration file change"""
    sigMechanicalFlipsChanged = qt.Signal()
    """Signal emitted when the mechanical flips have been changed"""

    def __init__(self, parent=None, host_discovery=None):
        qt.QMainWindow.__init__(self, parent)
        self.setWindowFlags(qt.Qt.Widget)
        self._listener = None
        self.lastFoundScans = LastReceivedScansDict(self.NB_STORED_LAST_FOUND)
        self._blissScans = {}
        # keep a trace of the bliss scans. key is bliss scan strings
        # (used as id), value is BlissScan instance
        self._syncWorkers = {}
        # associate scan path (directory) to the RSyncWorker

        # create widgets
        self._centralWidget = qt.QWidget(parent=self)
        self._centralWidget.setLayout(qt.QVBoxLayout())

        self._controlWidget = DataListenerControl(parent=self)
        """Widget containing the 'control' of the datalistener: start of stop
        the listener"""
        self._centralWidget.layout().addWidget(self._controlWidget)

        self._historyWindow = ScanHistory(parent=self)
        """Widget containing the latest valid scan found by the listener"""
        self._centralWidget.layout().addWidget(self._historyWindow)

        self._configWindow = ConfigurationWidget(
            parent=self, host_discovery=host_discovery
        )
        """Widget containing the configuration to communicate with bliss"""
        self._centralWidget.layout().addWidget(self._configWindow)

        self._observationWidget = ScanObservation(parent=self)
        """Widget containing the current observed directory by the listener"""
        self._centralWidget.layout().addWidget(self._observationWidget)

        # create toolbar
        toolbar = qt.QToolBar("")
        toolbar.setIconSize(qt.QSize(32, 32))

        self._controlAction = actions.ControlAction(parent=self)
        self._observationsAction = actions.ObservationAction(parent=self)
        self._configurationAction = actions.ConfigurationAction(parent=self)
        self._historyAction = actions.HistoryAction(parent=self)
        toolbar.addAction(self._controlAction)
        toolbar.addAction(self._observationsAction)
        toolbar.addAction(self._configurationAction)
        toolbar.addAction(self._historyAction)

        self._actionGroup = qt.QActionGroup(self)
        self._actionGroup.addAction(self._controlAction)
        self._actionGroup.addAction(self._observationsAction)
        self._actionGroup.addAction(self._configurationAction)
        self._actionGroup.addAction(self._historyAction)

        self.addToolBar(qt.Qt.LeftToolBarArea, toolbar)
        toolbar.setMovable(False)

        # signal / slot connection
        self._actionGroup.triggered.connect(self._updateCentralWidget)
        self._controlWidget.sigActivated.connect(self.sigActivate)
        self._controlWidget.sigDeactivated.connect(self.sigDeactivate)
        self._configWindow.sigConfigurationChanged.connect(self.sigConfigurationChanged)
        self._configWindow.sigCFGFileChanged.connect(self.sigCFGFileChanged)
        self._configWindow.sigMechanicalFlipsChanged.connect(
            self.sigMechanicalFlipsChanged
        )
        self._configWindow.sigRestart.connect(self._restartListener)

        # expose api
        self.activate = self._controlWidget.activate
        self.getCFGFilePath = self._configWindow.getCFGFilePath
        self.getOutputFolder = self._configWindow.getOutputFolder

        # set up
        self.setCentralWidget(self._centralWidget)
        self._controlAction.setChecked(True)
        self._updateCentralWidget(self._controlAction)

    def setHostAndPortToolTip(self, tooltip: str):
        self._configWindow._hostQLE.setToolTip(tooltip)
        self._configWindow._hostLabel.setToolTip(tooltip)
        self._configWindow._portLabel.setToolTip(tooltip)
        self._configWindow._portSpinBox.setToolTip(tooltip)

    def _updateCentralWidget(self, action_triggered):
        action_to_widget = {
            self._controlAction: self._controlWidget,
            self._historyAction: self._historyWindow,
            self._observationsAction: self._observationWidget,
            self._configurationAction: self._configWindow,
        }
        for action, widget in action_to_widget.items():
            widget.setVisible(action is action_triggered)

    def _serverStopped(self):
        self.sigServerStopped.emit()

    def _acquisitionStarted(self, arg: tuple):
        master_file, entry, proposal_file, saving_file = arg
        scan = self._getBlissScan(
            master_file=master_file, entry=entry, proposal_file=proposal_file
        )
        if settings.isOnLbsram(scan.path):
            self._attachRSyncWorker(scan.path, proposal_file, saving_file)
        self.addAcquisitionObserve(scan=scan)

    def _acquisitionEnded(self, arg: tuple):
        master_file, entry, proposal_file, saving_file, succeed = arg
        scan = self._getBlissScan(
            master_file=master_file, entry=entry, proposal_file=proposal_file
        )
        self.setAcquisitionEnded(scan=scan, success=succeed)
        if self._hasRSyncWorkerAttach(scan.path):
            self._detachRSyncWorker(scan.path)
        self.sigAcquisitionEnded.emit(
            (master_file, entry, proposal_file, saving_file, succeed)
        )

    def _acquisitionUpdated(self, arg: tuple):
        master_file, entry, proposal_file, saving_file, scan_number = arg
        scan = self._getBlissScan(
            master_file=master_file, entry=entry, proposal_file=proposal_file
        )
        scan.add_scan_number(scan_number)
        if settings.isOnLbsram(scan.path):
            if not self._hasRSyncWorkerAttach(scan.path):
                self._attachRSyncWorker(
                    scan.path, proposal_file=proposal_file, saving_file=saving_file
                )

        self.updateAcquisitionObserve(scan=scan)

    def _getBlissScan(self, master_file, entry, proposal_file):
        scan_id = BlissScan.get_id_name(master_file=master_file, entry=entry)
        if scan_id in self._blissScans:
            return self._blissScans[scan_id]
        else:
            bliss_scan = BlissScan(
                master_file=master_file, entry=entry, proposal_file=proposal_file
            )
            self._blissScans[str(bliss_scan)] = bliss_scan
            return bliss_scan

    def addAcquisitionObserve(self, scan):
        self._observationWidget.addObservation(scan)
        self._observationWidget.update(scan, "on going")

    def setAcquisitionEnded(self, scan, success):
        if success is False:
            self._observationWidget.update(scan, "failed")
        else:
            self._observationWidget.removeObservation(scan)
            self.lastFoundScans.add(scan)
            self._historyWindow.update(list(self.lastFoundScans.items()))

    def updateAcquisitionObserve(self, scan):
        self._observationWidget.update(scan, "on going")

    def sizeHint(self):
        return qt.QSize(600, 400)

    def _restartListener(self, host_name, host_port):
        # At the moment we still pass by the `BEACON_HOST` environment variable
        # Because the interface is still the same for the old data listener and the new one.
        # This is the most straightforward.
        os.environ["BEACON_HOST"] = f"{host_name}:{host_port}"
        self.activate(False)
        self.activate(True)

    def _attachRSyncWorker(self, scan_path, proposal_file, saving_file):
        dest_dir = scan_path.replace(
            settings.get_lbsram_path(), settings.get_dest_path()
        )
        dest_dir = os.path.dirname(dest_dir)
        if proposal_file is not None:
            dest_proposal_file = proposal_file.replace(
                settings.get_lbsram_path(), settings.get_dest_path()
            )
        else:
            dest_proposal_file = None
        if saving_file is not None:
            dest_saving_file = saving_file.replace(
                settings.get_lbsram_path(), settings.get_dest_path()
            )
        else:
            dest_saving_file = None
        worker = BlissSequenceRSyncWorker(
            src_dir=scan_path,
            dst_dir=dest_dir,
            delta_time=1,
            src_proposal_file=proposal_file,
            dst_proposal_file=dest_proposal_file,
            src_sample_file=saving_file,
            dst_sample_file=dest_saving_file,
        )
        self._syncWorkers[scan_path] = worker
        worker.start()

    def _detachRSyncWorker(self, scan_path):
        if self._hasRSyncWorkerAttach(scan_path=scan_path):
            worker = self._syncWorkers[scan_path]
            worker.stop()
            del self._syncWorkers[scan_path]

    def _hasRSyncWorkerAttach(self, scan_path):
        return scan_path in self._syncWorkers

    # expose API
    def getHost(self) -> str:
        """Return server host"""
        return self._configWindow.getHost()

    def getPort(self) -> int:
        """Return server port"""
        return self._configWindow.getPort()

    def getBlissServerConfiguration(self) -> dict:
        return self._configWindow.getConfiguration()

    def setBlissServerConfiguration(self, config):
        self._configWindow.setConfiguration(config=config)

    def setCFGFilePath(self, cfg_file):
        self._configWindow.setCFGFilePath(cfg_file)

    def setOutputFolder(self, output_dir):
        self._configWindow.setOutputFolder(output_dir)

    def setMechanicalFlips(self, left_right_flip: bool, up_down_flip: bool) -> None:
        self._configWindow.setMechanicalFlips(
            left_right_flip=left_right_flip, up_down_flip=up_down_flip
        )

    def getMechanicalFlips(self) -> tuple[bool, bool]:
        return self._configWindow.getMechanicalFlips()


class DataListenerControl(qt.QWidget):
    """Interface to control the activation of the datalistener"""

    sigActivated = qt.Signal()
    """signal emitted when the datalistener is start"""
    sigDeactivated = qt.Signal()
    """signal emitted when the datalistener is stop"""

    def __init__(self, parent=None):
        qt.QWidget.__init__(self, parent=parent)
        self.setLayout(qt.QGridLayout())

        # add left spacer
        lspacer = qt.QWidget(self)
        lspacer.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Minimum)
        self.layout().addWidget(lspacer, 0, 0, 1, 1)

        # add start / stop icon frame
        self._iconLabel = qt.QLabel(parent=self)
        self._iconLabel.setMinimumSize(qt.QSize(55, 55))
        self.layout().addWidget(self._iconLabel, 0, 1, 1, 1)

        # add button
        self._button = qt.QPushButton(self)
        self.layout().addWidget(self._button, 1, 1, 1, 1)

        # add right spacer
        rspacer = qt.QWidget(self)
        rspacer.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Minimum)
        self.layout().addWidget(rspacer, 0, 2, 1, 1)

        # bottom spacer
        bspacer = qt.QWidget(self)
        bspacer.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Expanding)
        self.layout().addWidget(bspacer, 2, 1, 1, 1)

        # set up
        self._updateIconAndText(activate=False)

        # connect signal / slot
        self._button.released.connect(self._buttonCallback)

    def _buttonCallback(self):
        self.activate(not self.isActivate())

    def isActivate(self):
        return self._button.text() == "stop"

    def activate(self, activate=True, *args, **kwargs):
        self._updateIconAndText(activate=activate)
        if activate is True:
            self.sigActivated.emit()
        else:
            self.sigDeactivated.emit()

    def _updateIconAndText(self, activate):
        if activate:
            icon = tomwericons.getQIcon("datalistener_activate")
        else:
            icon = tomwericons.getQIcon("datalistener_deactivate")

        text = "stop" if activate else "start"
        self._button.setText(text)
        self._iconLabel.setPixmap(icon.pixmap(80, 80))


class ConfigurationWidget(qt.QTabWidget):
    """Widget for data listener configuration"""

    sigConfigurationChanged = qt.Signal()
    """Signal emitted when the configuration change"""
    sigCFGFileChanged = qt.Signal(str)

    sigMechanicalFlipsChanged = qt.Signal()
    """Signal emitted when the mechanical flip has been changed"""

    sigRestart = qt.Signal(str, int)
    """Emit when use ask for a restart of the data listener. This can happen when the host / port are changed"""

    DEFAULT_PORT = 25_000

    def __init__(
        self,
        parent=None,
        host_discovery: str | None = None,
        enable_host_pinging: bool = True,
    ):
        """
        :param host_discovery: define a policy regarding host discovery when creating the widget. If None no discovery will be made.
            If 'BEACON_HOST' given then will try to discover the host name and port using the environment variable BEACON_HOST.
        """

        super().__init__(parent)

        # host tab
        self._hostWidget = qt.QDialog()
        self._hostWidget.setLayout(qt.QGridLayout())

        self._currentHost = self._getHost(host_discovery)
        """
        Current address listen by bliss data. As host, port.
        setPort and setHost will update this value.
        Else it will be updated when users are asking for a restart of the bliss-data listener.
        Used to handle the '_restartButton' and '_resetButton'
        """
        # host
        self._hostLabel = qt.QLabel("host name", self)
        self._hostWidget.layout().addWidget(self._hostLabel, 0, 0, 1, 1)
        self._hostQLE = HostEditor(
            parent=self,
            name=self._currentHost.name,
            port=self._currentHost.port,
            enable_host_pinging=enable_host_pinging,
        )
        self._hostQLE.setEnabled(self._hostEditable)
        self._hostWidget.layout().addWidget(self._hostQLE, 0, 1, 1, 2)

        # port
        self._portLabel = qt.QLabel("host port", self)
        self._hostWidget.layout().addWidget(self._portLabel, 1, 0, 1, 1)
        self._portSpinBox = qt.QSpinBox(self)
        self._portSpinBox.setRange(0, 100000)
        self._portSpinBox.setValue(self._currentHost.port)
        self._portSpinBox.setEnabled(self._hostEditable)
        self._hostWidget.layout().addWidget(self._portSpinBox, 1, 1, 1, 2)

        # restart & reset buttons
        self._buttons = qt.QDialogButtonBox(parent=self)
        style = qt.QApplication.style()
        icon = style.standardIcon(qt.QStyle.SP_MediaPlay)
        self._restartButton = qt.QPushButton(icon, "restart bliss-data listener")
        self._buttons.addButton(self._restartButton, qt.QDialogButtonBox.ActionRole)

        icon = style.standardIcon(qt.QStyle.SP_DialogResetButton)
        self._resetButton = qt.QPushButton(icon, "reset host")
        self._buttons.addButton(self._resetButton, qt.QDialogButtonBox.ActionRole)

        self._hostWidget.layout().addWidget(self._buttons, 2, 2, 1, 2)

        # spacer
        spacer = qt.QWidget(parent=self)
        spacer.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Expanding)
        self._hostWidget.layout().addWidget(spacer, 90, 0, 1, 1)

        # h52nx tab
        self._h52nxConfigWidget = qt.QWidget()
        self._h52nxConfigWidget.setLayout(qt.QGridLayout())

        # configuration file to use
        self._cfgLabel = qt.QLabel("config file")
        self._h52nxConfigWidget.layout().addWidget(self._cfgLabel, 11, 0, 1, 1)
        self._cfgWidget = HDF5ConfigFileSelector(self)
        self._cfgWidget.setContentsMargins(0, 0, 0, 0)
        self._cfgWidget.layout().setContentsMargins(0, 0, 0, 0)
        self._h52nxConfigWidget.layout().addWidget(self._cfgWidget, 11, 1, 1, 2)
        tooltip = (
            "You can provide a configuration file to tune conversion "
            "done by nxtomomill. If None is provided then the default "
            "parameters will be used."
        )
        self._cfgLabel.setToolTip(tooltip)
        self._cfgWidget.setToolTip(tooltip)

        # mechanical flip
        self._lrMechanicalFlipLabel = qt.QLabel("left-right mechanical flip")
        lr_icon = tomwericons.getQIcon("lr_mirroring")
        self._h52nxConfigWidget.layout().addWidget(
            self._lrMechanicalFlipLabel, 12, 0, 1, 1
        )
        self._lrMechanicalFlip = qt.QCheckBox()
        self._lrMechanicalFlip.setIcon(lr_icon)
        self._lrMechanicalFlip.setToolTip(
            "Detector image is flipped **left-right** for mechanical reasons that are not propagated with bliss-tomo hdf5 metadata (mirror,..)."
        )
        self._h52nxConfigWidget.layout().addWidget(self._lrMechanicalFlip, 12, 1, 1, 2)

        self._udMechanicalFlipLabel = qt.QLabel("up-down mechanical flip")
        ud_icon = tomwericons.getQIcon("ud_mirroring")
        self._h52nxConfigWidget.layout().addWidget(
            self._udMechanicalFlipLabel, 13, 0, 1, 1
        )
        self._udMechanicalFlip = qt.QCheckBox()
        self._udMechanicalFlip.setIcon(ud_icon)
        self._udMechanicalFlip.setToolTip(
            "Detector image is flipped **up-down** for mechanical reasons that are not propagated with bliss-tomo hdf5 metadata (mirror,..)."
        )
        self._h52nxConfigWidget.layout().addWidget(self._udMechanicalFlip, 13, 1, 1, 2)

        # vline 2
        self._vLine2 = qt.QFrame()
        self._vLine2.setFrameShape(qt.QFrame.Shape.HLine)
        self._vLine2.setFrameShadow(qt.QFrame.Shadow.Sunken)
        self._h52nxConfigWidget.layout().addWidget(self._vLine2, 20, 0, 1, 3)

        # output folder
        self._outputFolderLabel = qt.QLabel("nexus file output dir")
        self._h52nxConfigWidget.layout().addWidget(self._outputFolderLabel, 21, 0, 1, 1)
        self._nxTomomillOutputWidget = NXTomomillOutputDirSelector(self)
        self._nxTomomillOutputWidget.setContentsMargins(0, 0, 0, 0)
        self._nxTomomillOutputWidget.layout().setContentsMargins(0, 0, 0, 0)
        self._h52nxConfigWidget.layout().addWidget(
            self._nxTomomillOutputWidget, 21, 1, 1, 2
        )

        # buttons
        types = qt.QDialogButtonBox.Apply
        self._buttons = qt.QDialogButtonBox(self)
        self._buttons.setStandardButtons(types)
        self._buttons.button(qt.QDialogButtonBox.Apply).setToolTip(
            "Once apply if a listening is on going"
            "then it will stop the current listening and"
            "restart it with the new parameters"
        )
        self._h52nxConfigWidget.layout().addWidget(self._buttons, 21, 0, 1, 3)

        # height spacer
        spacer = qt.QWidget(parent=self)
        spacer.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Expanding)
        self._h52nxConfigWidget.layout().addWidget(spacer, 90, 0, 1, 1)

        # connect signal / slot
        self._cfgWidget.sigConfigFileChanged.connect(self.sigCFGFileChanged)
        self._buttons.button(qt.QDialogButtonBox.Apply).clicked.connect(self.validate)
        self._nxTomomillOutputWidget.sigChanged.connect(self.validate)
        self._lrMechanicalFlip.toggled.connect(self.sigMechanicalFlipsChanged)
        self._udMechanicalFlip.toggled.connect(self.sigMechanicalFlipsChanged)
        self._portSpinBox.valueChanged.connect(self._updateRestartAndResetAvailability)
        self._portSpinBox.editingFinished.connect(self._portHasChanged)
        self._hostQLE.currentTextChanged.connect(
            self._updateRestartAndResetAvailability
        )
        self._restartButton.released.connect(self._restartBlissData)
        self._resetButton.released.connect(self._resetHostAndPortValues)

        # set up
        self._buttons.hide()

        # update host name and port in case one was none at start up.
        self._currentHost.name = self.getHost()
        self._currentHost.port = self.getPort()
        self._updateRestartAndResetAvailability()

        self.addTab(self._hostWidget, "host")
        self.addTab(self._h52nxConfigWidget, "h52nx")

    def _getHost(self, host_discovery) -> Host:
        if settings.JSON_RPC_HOST is not None:
            # use case settings.JSON_RPC_HOST is defined
            # if defined in the settings file take it
            self._hostEditable = False
            return Host(settings.JSON_RPC_HOST, settings.JSON_RPC_PORT)

        if host_discovery is None:
            # use case settings.JSON_RPC_HOST is not defined but 'host_discovery' is None (requires JSON-RPC)
            self._hostEditable = False
            return Host(
                socket.gethostname(),
                settings.JSON_RPC_PORT,
            )

        # default use case - using bliss-data and BEACON_HOST
        self._hostEditable = True
        err_beacon_host = "Unable to determine host name and host port from 'BEACON_HOST'. Please export it as host_name:host_port. For example 'export BEACON_HOST=icc:0000'"
        beacon_host = os.environ.get("BEACON_HOST")
        if beacon_host is None:
            _logger.warning(err_beacon_host)
            return Host(None, self.DEFAULT_PORT)
        try:
            host_name, host_port = beacon_host.split(":")
            try:
                host_port = int(host_port)
            except ValueError:
                host_port = self.DEFAULT_PORT
        except ValueError:
            _logger.error(err_beacon_host)
            return Host(None, self.DEFAULT_PORT)
        else:
            return Host(host_name, host_port)

    def _restartBlissData(self):
        self._currentHost.name = self.getHost()
        self._currentHost.port = self.getPort()
        self.sigRestart.emit(self._currentHost.name, self._currentHost.port)
        self._updateRestartAndResetAvailability()

    def _resetHostAndPortValues(self):
        self.setHost(self._currentHost.name)
        self.setPort(self._currentHost.port)

    def _updateRestartAndResetAvailability(self):
        host_name_or_port_has_changed = (
            self._currentHost.name != self.getHost()
            or self._currentHost.port != self.getPort()
        )
        self._resetButton.setEnabled(host_name_or_port_has_changed)
        self._restartButton.setEnabled(host_name_or_port_has_changed)

    def _portHasChanged(self):
        # update port to the host (so it can update the icons). This requires a bit of processing so
        # we do this when edition is finished only.
        self._hostQLE.setPort(self.getPort())

    def getCFGFilePath(self):
        return self._cfgWidget.getCFGFilePath()

    def setCFGFilePath(self, cfg_file):
        self._cfgWidget.setCFGFilePath(cfg_file)

    def getOutputFolder(self):
        return self._nxTomomillOutputWidget.getOutputFolder()

    def setOutputFolder(self, output_dir):
        self._nxTomomillOutputWidget.setOutputFolder(output_dir)

    def addBlissSession(self, session: str) -> None:
        if self._blissSession.findText(session) >= 0:
            return
        else:
            self._blissSession.addItem(session)

    def getConfiguration(self) -> dict:
        return {"host": self.getHost(), "port": self.getPort()}

    def setConfiguration(self, config: dict):
        if "host" in config:
            self.setHost(config["host"])
        if "port" in config:
            self.setPort(config["port"])

    def getHost(self) -> str:
        return self._hostQLE.currentText()

    def setHost(self, name: str):
        self._hostQLE.setCurrentText(name)
        self._currentHost.name = name

    def getPort(self) -> int:
        return self._portSpinBox.value()

    def setPort(self, port: int) -> None:
        assert isinstance(port, int)
        self._portSpinBox.setValue(port)
        self._currentHost.port = port

    def getMechanicalFlips(self) -> tuple[bool, bool]:
        """
        :return: left-right flip, up-down flip
        """
        return self._lrMechanicalFlip.isChecked(), self._udMechanicalFlip.isChecked()

    def setMechanicalFlips(self, left_right_flip: bool, up_down_flip: bool) -> None:
        self._lrMechanicalFlip.setChecked(left_right_flip)
        self._udMechanicalFlip.setChecked(up_down_flip)

    def validate(self):
        self.sigConfigurationChanged.emit()



import pytest
from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt

from tomwer.gui.reconstruction.nabu.nabuconfig.ctf import CTFConfig
from tomwer.tests.utils import skip_gui_test


@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
class TestFutureSupervisorOW(TestCaseQt):
    def setUp(self):
        super().setUp()
        self._widget = CTFConfig()

    def tearDown(self):
        self._widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._widget.close()
        self._widget = None
        return super().tearDown()

    def test_configuration(self):
        """
        test configuration setter and getter
        """
        self._widget.show()
        config = self._widget.getConfiguration()
        assert isinstance(config, dict)

        assert config == {
            "ctf_geometry": " z1_v=None; z1_h=None; detec_pixel_size=None; magnification=True",
            "ctf_advanced_params": " length_scale=1e-05; lim1=1e-05; lim2=0.2; normalize_by_mean=True",
            "beam_shape": "parallel",
        }

        config["beam_shape"] = "cone"
        config["ctf_geometry"] = (
            " z1_v=10.2; z1_h=3.6; detec_pixel_size=1e-05; magnification=False"
        )
        config["ctf_advanced_params"] = (
            " length_scale=2e-05; lim1=1e-08; lim2=0.1; normalize_by_mean=False"
        )

        self._widget.setConfiguration(config)
        assert self._widget.getConfiguration() == config



"""
(Ewoks)Orange Widgets for pre or post processing stitching
"""

NAME = "stitching"

ID = "orangecontrib.tomwer.widgets.stitching"

DESCRIPTION = "widgets for data stitching"

LONG_DESCRIPTION = "Contains widget for stitching operation"

ICON = "../../widgets/icons/tomwer_stitching.png"

BACKGROUND = "#C0CCFF"

PRIORITY = 12



from nabu.stitching.utils import ShiftAlgorithm as _NabuShiftAlgorithm
from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt

from tomwer.gui.stitching.z_stitching.fineestimation import Axis_N_Params


class TestAxis_N_Params(TestCaseQt):
    """
    Test definition of an axis shift research parameters
    """

    def setUp(self):
        super().setUp()
        self._widget = Axis_N_Params("axis n", None)

    def tearDown(self):
        self._widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._widget.close()
        self._widget = None
        while self.qapp.hasPendingEvents():
            self.qapp.processEvents()

    def test(self):
        self._widget.show()
        assert self._widget.getCurrentMethod() == _NabuShiftAlgorithm.NONE
        opts1 = self._widget.getOptsLine()
        assert opts1 == "img_reg_method=None"
        self._widget.setCurrentMethod(_NabuShiftAlgorithm.NABU_FFT)
        opts2 = self._widget.getOptsLine()
        assert opts2 == "img_reg_method=nabu-fft"
        self._widget.setOptsLine(opts1)
        assert self._widget.getOptsLine() == opts1
        self._widget.setOptsLine(opts2)
        assert self._widget.getOptsLine() == opts2



"""
contains gui to select a slice in a volume
"""

from __future__ import annotations

import logging
from typing import Iterable

import numpy
from silx.gui import qt
from silx.gui.plot import PlotWindow
from silx.gui.plot.utils.axis import SyncAxes
from silx.io.url import DataUrl
from tomoscan.esrf.scan.utils import get_data
from math import floor

from tomwer.core.process.reconstruction.scores.scores import ComputedScore, ScoreMethod

_logger = logging.getLogger(__name__)


class VignettesQDialog(qt.QDialog):
    """ """

    SIZE_HINT = qt.QSize(820, 820)

    AUTO_NB_COLUMN = "auto"

    COLUMN_VALUES = (AUTO_NB_COLUMN, 1, 2, 3, 4, 5, 6, 8, 10)

    RESIZE_MAX_TIME = 500

    def __init__(
        self,
        value_name,
        score_name,
        parent=None,
        value_format=None,
        score_format=None,
        colormap=None,
    ):
        qt.QDialog.__init__(self, parent)
        self.setWindowFlags(qt.Qt.Widget)
        self._scan = None
        self._selectedValue = None
        self.setLayout(qt.QVBoxLayout())

        self._nbColumnCB = qt.QComboBox(self)
        for nb_column in self.COLUMN_VALUES:
            self._nbColumnCB.addItem(str(nb_column))

        self._columnWidget = qt.QWidget(self)
        # needed intermediary widget because the 'setWidgetResizable' fails with QFormLayout on PyQt 5.14.1 and will want to keep compatiblity for now
        self._columnWidget.setLayout(qt.QFormLayout())
        self._columnWidget.layout().addRow("number of column", self._nbColumnCB)
        self.layout().addWidget(self._columnWidget)

        self._mainWidget = qt.QScrollArea(self)
        self._mainWidget.setHorizontalScrollBarPolicy(qt.Qt.ScrollBarAlwaysOff)
        self._mainWidget.setVerticalScrollBarPolicy(qt.Qt.ScrollBarAsNeeded)
        self._vignettesWidget = VignettesWidget(
            self,
            with_spacer=True,
            value_name=value_name,
            score_name=score_name,
            score_format=score_format,
            value_format=value_format,
            colormap=colormap,
        )
        self._mainWidget.setWidget(self._vignettesWidget)
        self.layout().addWidget(self._mainWidget)
        self._mainWidget.setWidgetResizable(True)

        # buttons
        types = qt.QDialogButtonBox.Ok | qt.QDialogButtonBox.Cancel
        self._buttons = qt.QDialogButtonBox(parent=self)
        self._buttons.setStandardButtons(types)
        self._buttons.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Minimum)
        self.layout().addWidget(self._buttons)

        # connect signal slot
        self._buttons.button(qt.QDialogButtonBox.Ok).clicked.connect(self.accept)
        self._buttons.button(qt.QDialogButtonBox.Cancel).clicked.connect(self.reject)
        self._nbColumnCB.currentIndexChanged.connect(self._updateNbColumn)

        self._resizeCallback = None
        self._firstResizeEvent = True

    def sizeHint(self):
        """Return a reasonable default size for usage in :class:`PlotWindow`"""
        return self.SIZE_HINT

    def setScores(self, scores: dict, score_method: ScoreMethod):
        """
        Expect a dictionary with possible values to select as key and
        (2D numpy.array, score) as value.
        Where the 2D numpy.array is the frame to display and the score if the
        "indicator" score to display with the frame.
        :param scores: with score as key and a tuple (url|numpy array, ComputedScore) as value
        :param ScoreMethod score_method: score kind to be display
        """
        self._vignettesWidget.setScores(scores=scores, score_method=score_method)

    def selectedValue(self):
        if self._vignettesWidget is None:
            return self._selectedValue
        else:
            return self._vignettesWidget.selectedValue()

    def acccept(self):
        self._selectedValue = self._vignettesWidget.selectedValue()
        self._vignettesWidget.close()
        super().accept()

    def reject(self):
        self._selectedValue = self._vignettesWidget.selectedValue()
        self._vignettesWidget.close()
        super().reject()

    def getNColumn(self) -> int | None:
        n_column = self._nbColumnCB.currentText()
        if n_column == self.AUTO_NB_COLUMN:
            return None
        else:
            return int(n_column)

    def setNbColumn(self, value: int | str):
        if value not in self.COLUMN_VALUES:
            raise ValueError(
                f"Unhandled number of column requested ({value}). Valid values are {self.COLUMN_VALUES}"
            )
        idx = self._nbColumnCB.findText(str(value))
        self._nbColumnCB.setCurrentIndex(idx)

    @staticmethod
    def computeOptimalNColumn(window_width, vignette_width):
        n = floor(window_width / vignette_width)
        # To check with Jerome; He seems to have work on the topic.
        # might have some method for a smart way to do it
        return max(1, n)

    def _updateNbColumn(self, force_update=False):
        n_colum = self.getNColumn()
        if n_colum is None:
            n_colum = self.computeOptimalNColumn(
                window_width=self.width(),
                vignette_width=400,
            )
        self._vignettesWidget.setNElementsPerRow(n_colum, force_update=force_update)

    def resizeEvent(self, event):
        if self._firstResizeEvent:
            # cheap way to filter the first resize and avoid updating it
            self._firstResizeEvent = False
        else:
            # print("resize event received...", event)
            if self._resizeCallback is not None:
                self._resizeCallback.timeout.disconnect(self._updateNbColumn)
                self._resizeCallback.stop()
                self._resizeCallback.deleteLater()
                self._resizeCallback = None
            self._resizeCallback = qt.QTimer(self)
            self._resizeCallback.setSingleShot(True)
            self._resizeCallback.timeout.connect(self._updateNbColumn)
            self._resizeCallback.start(self.RESIZE_MAX_TIME)

        super().resizeEvent(event)


class VignettesWidget(qt.QWidget):
    """
    Widget to display all the frames.

    :param parent:
    :param value_name: name of the values for which we are looking for the
                           best one
    :param score_name: name of the score computed
    :param score_format: None or str that can be formatted to display the
                         score
    :param value_format: None or str that can be formatted to display the
                         value
    """

    DEFAULT_PLOT_PER_ROW = 2
    MAX_NB_COLUMN = 999999

    def __init__(
        self,
        parent=None,
        value_name="value",
        score_name="score",
        with_spacer=True,
        value_format=None,
        score_format=None,
        colormap=None,
    ):
        qt.QWidget.__init__(self, parent)
        self._valueName = value_name
        self._scoreName = score_name
        self._nPlotPerRow = VignettesWidget.DEFAULT_PLOT_PER_ROW
        self._withSpacer = with_spacer
        self._vignettesGroup = qt.QButtonGroup()
        self.__constraintXAxis = None
        self.__constraintYAxis = None
        self._valueFormat = value_format
        self._scoreFormat = score_format
        self._colormap = colormap
        self._vignettes = []
        self.__score_method = None

        self.setLayout(qt.QGridLayout())

    def close(self):
        for vignette in self._vignettes:
            vignette.close()
        super().close()

    def selectedValue(self):
        sel_vignette = self._vignettesGroup.checkedButton()
        if sel_vignette is not None:
            return sel_vignette.getScoreValue()
        else:
            return None

    def setNElementsPerRow(self, n: int, force_update):
        assert isinstance(n, int), "number of column must be an int"
        if self._nPlotPerRow != n:
            self._nPlotPerRow = n
            self.__update()
        elif force_update:
            self.__update()

    def __update(self):
        scores, score_method = self.getScores()
        self.setScores(scores, score_method)
        # raise NotImplementedError

    def getNElementsPerRow(self):
        return self._nPlotPerRow

    def getScores(self) -> tuple:
        """
        Return currently displayed scores as (scores, scores_method)
        """
        scores = {}
        scores_method = self.__score_method
        for vignette in self._vignettes:
            assert isinstance(vignette, Vignette)
            if scores_method in (ScoreMethod.STD, ScoreMethod.STD_INVERSE):
                compute_score_args = {
                    "std": vignette.getScoreValue(),
                    "tv": None,
                }
            elif scores_method in (ScoreMethod.TV, ScoreMethod.TV_INVERSE):
                compute_score_args = {
                    "std": None,
                    "tv": vignette.getScoreValue(),
                }
            elif scores_method in (ScoreMethod.TOMO_CONSISTENCY):
                compute_score_args = {
                    "std": None,
                    "tv": None,
                    "tomo_consitency": vignette.getScoreValue(),
                }
            else:
                raise ValueError(f"score method unhandled: {scores_method.value}")

            scores[vignette.getValue()] = (
                vignette.getData(),
                ComputedScore(**compute_score_args),
            )
        return scores, scores_method

    def clearVignettes(self):
        for vignette in self._vignettes:
            self.layout().removeWidget(vignette)
            vignette.setParent(None)
            self._vignettesGroup.removeButton(vignette)
            # vignette.deleteLater()
        self._vignettes = []

    def setScores(self, scores: dict, score_method: ScoreMethod):
        """
        Expect a dictionary with possible values to select as key and
        (2D numpy.array, score) as value.
        Where the 2D numpy.array is the frame to display and the score if the
        "indicator" score to display with the frame.
        :param scores: with score as key and a tuple (url|numpy array, ComputedScore) as value
        :param ScoreMethod score_method: score kind to be display
        """
        if len(scores) < 1:
            return

        self.clearVignettes()

        i_row = 0
        scores_values = []
        for i_score, (value, (data, score_cls)) in enumerate(scores.items()):
            if not isinstance(score_cls, ComputedScore):
                raise TypeError(
                    f"score is expected to be a dict with values as (v1: numpy.ndarray, v2: ComputedScore). v2 type Found: {type(score_cls)}"
                )
            scores_values.append(score_cls.get(score_method))
        self.__score_method = ScoreMethod(score_method)
        highest_score_indices = numpy.nanargmax(scores_values)
        self._vignettesGroup = qt.QButtonGroup(self)
        self._vignettesGroup.setExclusive(True)

        if not isinstance(highest_score_indices, Iterable):
            highest_score_indices = (highest_score_indices,)

        xAxis = []
        yAxis = []

        for i_score, (value, (data, score_cls)) in enumerate(scores.items()):
            score = score_cls.get(score_method)
            i_column = i_score % self.getNElementsPerRow()
            # TODO: instead of having a binary color we could use
            # colormap from green (good score) to red (bad score score)
            if i_score == highest_score_indices or i_score in highest_score_indices:
                frame_color = qt.Qt.green
            else:
                frame_color = qt.Qt.lightGray
            widget = Vignette(
                parent=self,
                value_name=self._valueName,
                score_name=self._scoreName,
                value=value,
                data=data,
                score=score,
                frame_color=frame_color,
                value_format=self._valueFormat,
                score_format=self._scoreFormat,
                colormap=self._colormap,
            )
            widget.setAttribute(qt.Qt.WA_DeleteOnClose)
            xAxis.append(widget.getPlotWidget().getXAxis())
            yAxis.append(widget.getPlotWidget().getYAxis())

            self.layout().addWidget(widget, i_row, i_column)
            if i_column == self.getNElementsPerRow() - 1:
                i_row += 1
            if i_score == 0:
                # we cannot request all widget to keep the aspect ratio
                widget.setKeepDataAspectRatio(True)
            self._vignettesGroup.addButton(widget)
            self._vignettes.append(widget)

        if self._withSpacer:
            spacer = qt.QWidget(self)
            spacer.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Expanding)
            # to simplify we add a spacer at the end. We do not expect to have more than 999999 column
            self.layout().addWidget(spacer, i_row + 1, self.MAX_NB_COLUMN)

        # constrain axis synchronization
        self.__constraintXAxis = SyncAxes(
            xAxis,
            syncLimits=False,
            syncScale=True,
            syncDirection=True,
            syncCenter=True,
            syncZoom=True,
        )
        self.__constraintYAxis = SyncAxes(
            yAxis,
            syncLimits=False,
            syncScale=True,
            syncDirection=True,
            syncCenter=True,
            syncZoom=True,
        )


class _PlotForVignette(PlotWindow):
    def __init__(self, parent=None):
        PlotWindow.__init__(
            self,
            parent=parent,
            yInverted=True,
            copy=False,
            save=False,
            print_=False,
            control=False,
            mask=False,
        )
        self.setKeepDataAspectRatio(False)
        self.setAxesDisplayed(False)
        self.toolBar().hide()
        self.getInteractiveModeToolBar().hide()
        self.getOutputToolBar().hide()
        self.setInteractiveMode("zoom", zoomOnWheel=False)

    def close(self) -> bool:
        super().close()


class Vignette(qt.QToolButton):
    """Widget to display a vignette"""

    FRAME_WIDTH = 2

    def __init__(
        self,
        parent,
        value,
        value_name: str,
        score_name: str,
        data: DataUrl | numpy.array,
        score: float,
        frame_color: qt.QColor,
        score_format=None,
        value_format=None,
        colormap=None,
    ):
        self._value = value
        self._scoreName = score_name
        self._valueName = value_name
        qt.QToolButton.__init__(self, parent)
        self.setCheckable(True)
        self.setLayout(qt.QVBoxLayout())
        self._plot = _PlotForVignette(parent=self)
        self._plot.setDefaultColormap(colormap=colormap)

        self.layout().addWidget(self._plot)
        self._valueLabel = ValueLabel(
            self,
            value=value,
            score=score,
            score_name=self._scoreName,
            value_name=self._valueName,
            value_format=value_format,
            score_format=score_format,
        )
        self.layout().addWidget(self._valueLabel)
        self.setFixedSize(400, 400)
        self._frameColor = frame_color
        self._selectedFrameColor = qt.Qt.black

        if isinstance(data, DataUrl):
            data = get_data(data)
        if data.ndim == 3 and data.shape[0] == 1:
            data = data.reshape(data.shape[1:])
        self._plot.addImage(data)

    def setKeepDataAspectRatio(self, keep):
        self._plot.setKeepDataAspectRatio(keep)

    def getPlotWidget(self):
        return self._plot

    def getValue(self):
        return self._value

    def getData(self) -> numpy.ndarray:
        return self._plot.getImage().getData()

    def getScoreValue(self):
        return self._valueLabel.score

    def getScoreName(self):
        return self._scoreName

    def getValueName(self):
        return self._valueName

    def paintEvent(self, event):
        super().paintEvent(event)
        painter = qt.QPainter(self)
        half_h_width = self.FRAME_WIDTH // 2
        rect = qt.QRect(
            half_h_width,
            half_h_width,
            self.width() - self.FRAME_WIDTH,
            self.height() - self.FRAME_WIDTH,
        )
        pen = qt.QPen()
        pen.setWidth(Vignette.FRAME_WIDTH)
        pen.setColor(self._frameColor)
        painter.setPen(pen)
        painter.drawRect(rect)
        if self.isChecked():
            pen.setColor(self._selectedFrameColor)
            pen.setStyle(qt.Qt.DashLine)
            pen.setDashOffset(0.2)
            painter.setPen(pen)
            painter.drawRect(rect)

    def close(self):
        self._plot.clear()
        self._plot.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._plot.close()
        self._plot = None


class ValueLabel(qt.QWidget):
    """Display the value and the associated score"""

    def __init__(
        self, parent, value, score, value_name, score_name, value_format, score_format
    ):
        qt.QWidget.__init__(self, parent)
        self._score = score
        self.setLayout(qt.QHBoxLayout())
        if value_format is not None:
            str_value = value_format.format(value)
        else:
            str_value = str(value)
        txt = f"{value_name}: {str_value}"
        self._valueLabel = qt.QLabel(txt, self)
        self.layout().addWidget(self._valueLabel)
        if score_format is not None:
            str_score = score_format.format(score)
        else:
            str_score = str(score)
        txt = f"({score_name}: {str_score})"
        self._scoreLabel = qt.QLabel(txt, self)
        self.layout().addWidget(self._scoreLabel)

    @property
    def score(self):
        return self._score



"""contains 'FutureTomwerObject'. Class used when computation is done asynchronously (when submitted to slurm for example)"""

from __future__ import annotations

import datetime
import logging
from typing import Iterable

from sluurp.job import get_job_status, cancel_slurm_job

from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.tomwer_object import TomwerObject

_logger = logging.getLogger(__name__)


class FutureTomwerObject:
    """Container for an existing TomwerScanBase with some future object
    (like slice reconstruction on slurm...)

    :param scan: scan at the origin of the FutureTomwerScan
    :param futures: tuple of Future
    :param start_time: start_time of the FutureTomwerScan. If not provided
                       this will be set to the date of creation of the object.
    :param process_requester_id: process of the id requesting future
                                     processing. This is useful to keep a trace on
                                     the source of the computation. To display
                                     supervisor 'request' for example.
    """

    def __init__(
        self,
        tomo_obj: TomwerObject,
        futures: tuple | list | None,
        start_time: datetime.time | None = None,
        process_requester_id: int | None = None,
    ):
        if not isinstance(tomo_obj, TomwerObject):
            raise TypeError(
                f"tomo_obj is expected to be an instance of {TomwerObject} not {type(tomo_obj)}"
            )
        self._job_id = None
        self._tomo_obj = tomo_obj
        self._start_time = (
            start_time if start_time is not None else datetime.datetime.now()
        )
        self._process_requester_id = process_requester_id
        # handle futures
        if futures is None:
            self._futures = None
        elif not isinstance(futures, (tuple, list)):
            raise TypeError(
                f"futures is expected to be a tuple or a list not {type(futures)}"
            )
        else:
            self._futures = list(futures)

    @property
    def job_id(self) -> int | None:
        return self._job_id

    @property
    def status(self) -> str | None:
        all_status = [
            get_job_status(fut.job_id) if hasattr(fut, "job_id") else "unknown"
            for fut in self._futures
        ]
        # small hack to get a common status for a list of future.
        # if at least one running we can wait for all to be processed
        # Then after this condition it means all jobs are finished
        # and we can check if there is an error
        # otherwise everything is "green"
        for status in (
            "pending",
            "running",
            "error",
            "cancelled",
            "finished",
            "completed",
        ):
            if status in all_status:
                return status
        return "finished"

    @property
    def logs(self) -> str:
        new_line = "\n"
        return "\n".join(
            [
                f"~~~~~~ job id = {fut.job_id} ~~~~~{new_line}{f'{new_line}'.join(fut.collect_logs() or ['',])}"
                for fut in self._futures
            ]
        )

    @property
    def start_time(self) -> datetime.time | None:
        return self._start_time

    @property
    def tomo_obj(self) -> TomwerScanBase:
        # TODO: replace the scan by the DatasetIdentifier
        return self._tomo_obj

    @property
    def futures(self) -> tuple | None:
        if self._futures is None:
            return None
        else:
            return tuple(self._futures)

    def extend_futures(self, futures: Iterable) -> None:
        if not isinstance(futures, Iterable):
            raise TypeError(f"futures should be an Iterable not {type(futures)}")
        if self._futures is None:
            self._futures = []
        self._futures.extend(futures)

    @property
    def process_requester_id(self) -> int | None:
        return self._process_requester_id

    def results(self) -> TomwerScanBase:
        """Force future and callback synchronization"""
        results = []
        [results.append(fut.result()) for fut in self._futures]
        return results

    def cancelled(self):
        cancelled = False
        for future in self.futures:
            cancelled = cancelled and future.cancelled()
        return cancelled

    def done(self):
        done = True
        for future in self.futures:
            done = done and future.done()
        return done

    def exceptions(self) -> dict | None:
        """
        Return exception with future as key and exception as value
        """
        exceptions = {}
        for future in self.futures:
            if future.exception():
                exceptions[future] = future.exception()

        if len(exceptions) == 0:
            return None
        else:
            return exceptions

    def cancel(self):
        for future in self.futures:
            if hasattr(future, "job_id"):
                print("cancel ", future.job_id)
                cancel_slurm_job(future.job_id)



# coding: utf-8
from __future__ import annotations

import copy
import logging

import nxtomomill.version
from nxtomomill.utils import change_image_key_control as _change_image_key_control
from nxtomo.nxobject.nxdetector import ImageKey
from tomwer.core.utils.deprecation import deprecated_warning, deprecated

from tomwer.core.process.task import Task
from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.utils.scanutils import data_identifier_to_scan

_logger = logging.getLogger(__name__)


IMAGE_KEYS = {
    "projection": ImageKey.PROJECTION,
    "invalid": ImageKey.INVALID,
    "dark": ImageKey.DARK_FIELD,
    "flat": ImageKey.FLAT_FIELD,
}


def change_image_key_control(scan: NXtomoScan, config: dict) -> TomwerScanBase:
    """

    :param scan:
    :param config:
    :raises KeyError: if 'frames_indexes' or 'image_key_control_value' are
                      not in config
    :return:
    """
    if scan is None:
        return
    elif not isinstance(scan, NXtomoScan):
        raise ValueError(
            f"Image key control only handle NXtomoScan and not {type(scan)}"
        )

    if "modifications" not in config:
        raise KeyError("modifications are not provided")
    else:
        modifications = config["modifications"]
        if modifications is None:
            modifications = {}

    image_keys_set = set(modifications.values())
    image_keys_set = set([ImageKey(image_key) for image_key in image_keys_set])
    for image_key_type in image_keys_set:
        frame_indexes_dict = dict(
            filter(lambda item: item[1] is image_key_type, modifications.items())
        )
        frame_indexes = tuple(frame_indexes_dict.keys())
        _logger.info(f"will modify {frame_indexes} to {image_key_type}")
        _change_image_key_control(
            file_path=scan.master_file,
            entry=scan.entry,
            frames_indexes=frame_indexes,
            image_key_control_value=image_key_type.value,
            logger=_logger,
        )
    scan.clear_frames_cache()
    return scan


class ImageKeyEditorTask(
    Task,
    input_names=("data", "configuration"),
    output_names=("data",),
    optional_input_names=("serialize_output_data",),
):
    """
    task to edit `image_key` field of a NXtomo ('data' input)
    """

    def __init__(
        self, varinfo=None, inputs=None, node_id=None, node_attrs=None, execinfo=None
    ):
        super().__init__(
            varinfo=varinfo,
            inputs=inputs,
            node_id=node_id,
            node_attrs=node_attrs,
            execinfo=execinfo,
        )

    def run(self):
        scan = data_identifier_to_scan(self.inputs.data)
        if scan is None:
            return
        if not isinstance(scan, TomwerScanBase):
            raise TypeError(
                f"scan is expected to be a dict or an instance of TomwerScanBase. Not {type(scan)}"
            )
        if not isinstance(scan, NXtomoScan):
            raise ValueError(f"input type of {scan}: {(type(scan))} is not managed")

        config = self.inputs.configuration
        if not isinstance(config, dict):
            raise TypeError
        # apply configuration
        change_image_key_control(scan=scan, config=config)
        modif_keys = list(config.get("modifications", {}).keys())
        # dump modifications to the tomwer processes file
        # note: we need to cast image keys to str to save it
        config = copy.deepcopy(config)
        new_modif = {}
        for key in modif_keys:
            value = config["modifications"][key]
            config["modifications"][str(key)] = value
        config["modifications"] = new_modif

        if self.get_input_value("serialize_output_data", True):
            self.outputs.data = scan.to_dict()
        else:
            self.outputs.data = scan

    @staticmethod
    def program_name():
        return "nxtomomill.utils.change_image_key_control"

    @staticmethod
    def program_version():
        return nxtomomill.version.version

    @staticmethod
    def definition():
        return "Modify image keys on a NXtomo"


class ImageKeyUpgraderTask(
    Task,
    input_names=("data", "operations"),
    output_names=("data",),
    optional_input_names=("serialize_output_data",),
):
    """
    close to ImageKeyEditor but convert a full "family" of frame type to another
    like all projections to dark field
    """

    def run(self):
        scan = data_identifier_to_scan(self.inputs.data)
        if not isinstance(scan, NXtomoScan):
            raise TypeError(f"scan is expected to be an instance of {NXtomoScan}")
        operations = self.inputs.operations
        if not isinstance(operations, dict):
            raise TypeError("operations is expected to be an dict")

        configuration = {}
        for key_image_value_from, key_image_value_to in operations.items():
            configuration.update(
                self.from_operation_to_config(
                    scan=scan,
                    from_image_key=key_image_value_from,
                    to_image_key=key_image_value_to,
                )
            )
        configuration = {
            "modifications": configuration,
        }

        # apply modification using tomoscan
        change_image_key_control(scan=scan, config=configuration)

        if self.get_input_value("serialize_output_data", True):
            self.outputs.data = scan.to_dict()
        else:
            self.outputs.data = scan

    @staticmethod
    def from_operation_to_config(
        scan: TomwerScanBase, from_image_key: ImageKey, to_image_key: ImageKey
    ):
        """
        retrieve frame indices to be updated
        """
        from_image_key = ImageKey(from_image_key)
        to_image_key = ImageKey(to_image_key)

        config = {}
        for i_frame, frame in enumerate(scan.frames):
            if frame.image_key == from_image_key:
                config[i_frame] = to_image_key

        return config

    @staticmethod
    def program_name():
        return "nxtomomill.utils.ImageKeyUpgraderTask"

    @staticmethod
    def program_version():
        return nxtomomill.version.version

    @staticmethod
    def definition():
        return "Modify image keys on a NXtomo"

    @deprecated(
        since_version="1.2",
        replacement="DarkFlatPatchTask.inputs.configuration",
        reason="ewoksification",
    )
    def get_configuration(self):
        """

        :return: configuration of the process
        """
        return self.inputs.configuration

    @deprecated(
        since_version="1.2",
        replacement="DarkFlatPatchTask.inputs.configuration",
        reason="ewoksification",
    )
    def set_configuration(self, configuration: dict) -> None:
        self.inputs.configuration = configuration


class ImageKeyUpgrader(ImageKeyUpgraderTask):
    def __init__(
        self, varinfo=None, inputs=None, node_id=None, node_attrs=None, execinfo=None
    ):
        deprecated_warning(
            name="tomwer.core.process.edit.imagekeyeditor.ImageKeyUpgrader",
            type_="class",
            reason="improve readibility",
            since_version="1.2",
            replacement="ImageKeyUpgraderTask",
        )
        super().__init__(varinfo, inputs, node_id, node_attrs, execinfo)



from __future__ import annotations

from silx.gui import qt
from contextlib import contextmanager


@contextmanager
def qitem_model_resetter(model: qt.QAbstractItemModel):
    """Context manager for resetting a QAbstractItemModel.

    Make sure `beginResetModel()` is emit at the beginning and endResetModel() at the end.
    """
    model.beginResetModel()
    try:
        yield
    finally:
        model.endResetModel()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Application to stitch along z (aka axis 0) a set of scans or reconstructed volumes
"""
from __future__ import annotations

import sys
import logging
from tomwer.app.stitching.common import main as common_main

logging.basicConfig(level=logging.WARNING)

_logger = logging.getLogger(__name__)


def getinputinfo():
    return "tomwer z-stitching [tomo objects]"


def main(argv):
    common_main(argv, stitcher_name="z-stitching", stitching_axis=0, logger=_logger)


if __name__ == "__main__":
    main(sys.argv)



# coding: utf-8
from __future__ import annotations


import os

import pytest
from nxtomo.nxobject.nxdetector import ImageKey

from tomwer.core.scan.edfscan import EDFTomoScan
from tomwer.core.utils.scanutils import MockEDF, MockNXtomo
from tomwer.gui.reconstruction.nabu import check


def test_check_dark_series(tmpdir):
    """test check.check_dark_series function"""
    full_edf_path = os.path.join(tmpdir, "my", "aquisition", "folder")
    MockEDF.fastMockAcquisition(full_edf_path)
    edf_scan = EDFTomoScan(full_edf_path)
    with pytest.raises(TypeError):
        assert check.check_dark_series(edf_scan)

    full_hdf5_scan = os.path.join(tmpdir, "hdf5_scan")
    scan = MockNXtomo(
        scan_path=full_hdf5_scan,
        n_proj=20,
        n_ini_proj=20,
        dim=10,
    ).scan
    scan._image_keys = (
        [ImageKey.DARK_FIELD] * 2
        + [ImageKey.PROJECTION] * 4
        + [ImageKey.DARK_FIELD] * 2
    )
    assert check.check_dark_series(scan, logger=None, user_input=False) is False
    scan._image_keys = [ImageKey.DARK_FIELD] * 2 + [ImageKey.PROJECTION] * 4
    assert check.check_dark_series(scan, logger=None, user_input=False) is True
    scan._image_keys = [ImageKey.PROJECTION] * 4
    assert check.check_dark_series(scan, logger=None, user_input=False) is False


def test_check_flat_series(tmpdir):
    """test check.check_flat_series function"""
    full_edf_path = os.path.join(tmpdir, "my", "aquisition", "folder")
    MockEDF.fastMockAcquisition(full_edf_path)
    edf_scan = EDFTomoScan(full_edf_path)
    with pytest.raises(TypeError):
        assert check.check_flat_series(edf_scan)

    full_hdf5_scan = os.path.join(tmpdir, "hdf5_scan")
    scan = MockNXtomo(
        scan_path=full_hdf5_scan,
        n_proj=20,
        n_ini_proj=20,
        dim=10,
    ).scan
    scan._image_keys = (
        [ImageKey.FLAT_FIELD] * 2
        + [ImageKey.PROJECTION] * 4
        + [ImageKey.FLAT_FIELD] * 2
    )
    assert check.check_flat_series(scan, logger=None, user_input=False) is True
    scan._image_keys = [ImageKey.FLAT_FIELD] * 2 + [ImageKey.PROJECTION] * 4
    assert check.check_flat_series(scan, logger=None, user_input=False) is True
    scan._image_keys = [ImageKey.PROJECTION] * 4
    assert check.check_flat_series(scan, logger=None, user_input=False) is False



import os
from tomwer.core.process.reconstruction.params_cache import (
    _CACHE,
    _CACHE_SIZE,
    save_reconstruction_parameters_to_cache,
    clear_reconstruction_parameters_cache,
)
from tomwer.core.utils.scanutils import MockNXtomo


def test_cache(tmp_path):
    nxtomos = [
        MockNXtomo(
            scan_path=os.path.join(tmp_path, f"my_nxtomo_{i_scan}"),
            n_proj=10,
        ).scan
        for i_scan in range(31)
    ]
    new_nxtomo = MockNXtomo(
        scan_path=os.path.join(tmp_path, "my_nxtomo_99"),
        n_proj=10,
    ).scan

    clear_reconstruction_parameters_cache()
    # test saving it in the original order
    [save_reconstruction_parameters_to_cache(scan=scan) for scan in nxtomos]
    assert len(_CACHE) == _CACHE_SIZE
    assert _CACHE == {
        nxtomo.get_identifier().to_str(): (None, None) for nxtomo in nxtomos[1:]
    }

    # try updating the next last one to make sure it won't be removed
    save_reconstruction_parameters_to_cache(scan=nxtomos[1])
    save_reconstruction_parameters_to_cache(scan=new_nxtomo)
    assert nxtomos[1].get_identifier().to_str() in _CACHE
    assert nxtomos[2].get_identifier().to_str() not in _CACHE
    clear_reconstruction_parameters_cache()



from tomwer.core.process.stitching.metadataholder import StitchingMetadata


def test_to_dict_from_dict():
    """Test conversion of StitchingMetadata from dict and to dict"""
    metadata = StitchingMetadata(tomo_obj=None)
    metadata._pixel_or_voxel_size = [12.3, 2.5, None]
    metadata._pos_as_m = [None, 3.6, None]
    metadata._pos_as_px = [5, 9, 8]

    ddict = metadata.to_dict()
    assert isinstance(ddict, dict)

    loaded_metadata = StitchingMetadata.from_dict(ddict, tomo_obj=None)
    assert loaded_metadata == metadata
    assert loaded_metadata.to_dict() == ddict
    assert isinstance(str(loaded_metadata), str)



import numpy

from orangecontrib.tomwer.widgets.reconstruction.NabuVolumeOW import (
    _DeltaBetaSelectorDialog,
)
from tomwer.tests.conftest import qtapp  # noqa F401


def test_DeltaBetaSelector(
    qtapp,  # noqa F811
):
    """simple test of the _DeltaBetaSelectorDialog"""
    dialog = _DeltaBetaSelectorDialog(values=(12, 45))
    dialog.show()
    assert numpy.isscalar(dialog.getSelectedValue())



from __future__ import annotations


import logging
import os

import numpy.lib.npyio
from silx.gui import qt

import tomwer.resources
from tomwer.io.utils import get_slice_data

logger = logging.getLogger(__name__)


class ImageLoaderThread(qt.QThread):
    """Thread used to load an image"""

    IMG_NOT_FOUND = numpy.load(
        tomwer.resources._resource_filename(
            "%s.%s" % ("imageNotFound", "npy"),
            default_directory=os.path.join("gui", "icons"),
        )
    )

    def __init__(self, url, *args, **kwargs):
        """

        :param index: index of the image on the stackplot
        :param filePath: filePath is the file to load on stackplot reference.
                         It can be an .edf file or a .vol file. If this is a
                         vol file then the name is given with the slice z index
                         to be loaded.
        """
        super().__init__(*args, **kwargs)
        self.data = None
        self.url = url

    def getData(self):
        if hasattr(self, "data"):
            return self.data
        else:
            return None

    def run(self):
        if os.path.exists(self.url.file_path()) and os.path.isfile(
            self.url.file_path()
        ):
            self.data = get_slice_data(self.url)
        else:
            logger.warning("file %s not longer exists or is empty" % self.url)
            self.data = self.IMG_NOT_FOUND



from __future__ import annotations

import logging
from orangecontrib.tomwer.orange.managedprocess import TomwerUnsupervisedOneThreadPerRun

from orangewidget import gui

from ewokscore.missing_data import is_missing_data

from tomwer.tasks.visualization.volume_viewer import VolumeViewerTask
from tomwer.gui.visualization.volume_viewer.VolumeViewerWindow import (
    VolumeViewerWindow,
)


_logger = logging.getLogger(__name__)


class VolumeViewerOW(
    TomwerUnsupervisedOneThreadPerRun,
    ewokstaskclass=VolumeViewerTask,
):
    """
    Widget embedding a viewer to display reconstructed volume.

    Several views are possible. One per axis or an 'overall' view (displaying the three axis in parallel).

    By default this widget allow users to browse the volume along the 'fast axis' and will sample 3 slices along each axis that could be displayed.
    If users agrees on loading the full volume then browsing can be done along any direction.
    """

    name = "volume viewer"
    id = "orange.widgets.tomwer.visualization.VolumeViewerOW.VolumeViewerOW"
    description = "Widget for browsing reconstructed volume."

    icon = "icons/volumeviewer.svg"
    priority = 31
    keywords = [
        "tomography",
        "volume",
        "tomwer",
        "reconstruction",
        "summary",
        "slices",
        "axis",
    ]

    _ewoks_inputs_to_hide_from_orange = ("load_volume",)

    want_main_area = True
    want_control_area = False
    resizing_enabled = True

    def __init__(self, *args, **kwargs):
        super().__init__(self, *args, **kwargs)
        self._window = VolumeViewerWindow()
        self._box = gui.vBox(self.mainArea, self.name)
        self._box.layout().addWidget(self._window)
        self.loadSettings()

        # make sure 'load_volume' is defined.
        self.set_default_input("load_volume", self._window.isLoadingVolume())

        self._window.sigLoadVolume.connect(self._loadVolumeHasChanged)

    def handleNewSignals(self):
        volume = self.get_task_input_value("volume")
        if not is_missing_data(volume):
            self._startProcessing()
            self._window.initVolumePreview(volume=volume, message="Loading slices")

        super().handleNewSignals()

    def _loadVolumeHasChanged(self, load: bool):
        self.set_default_input("load_volume", load)

    def loadSettings(self):
        load_volume = self.get_default_input_value("load_volume", None)
        if load_volume is not None:
            self._window.setLoadingVolume(loading_volume=load_volume)

    def task_output_changed(self):
        slices = self.get_task_output_value("slices")
        volume_metadata = self.get_task_output_value("volume_metadata")
        volume_shape = self.get_task_output_value("volume_shape")
        loaded_volume = self.get_task_output_value("loaded_volume")

        if (not is_missing_data(slices)) and (not is_missing_data(volume_metadata)):
            self._window.setSlicesAndMetadata(
                slices=slices,
                metadata=volume_metadata,
                volume_shape=volume_shape,
            )

            if not is_missing_data(loaded_volume) and loaded_volume is not None:
                self._window.setVolume(loaded_volume)
        else:
            _logger.warning("Fail to load slices / metadata")
        self._window.setLoading(False)

        self._endProcessing()
        super().task_output_changed()



# coding: utf-8
from __future__ import annotations

from tomwer.core.process.reconstruction.darkref.darkrefs import (
    requires_reduced_dark_and_flat,
)
from tomwer.core.utils.scanutils import MockEDF, MockNXtomo


def test_quick_run_necessary_edf(tmpdir):
    """test the `quick_run_necessary` function for EDFTomoScan"""
    scan = MockEDF.mockScan(scanID=str(tmpdir), start_dark=True, start_flat=True)
    assert scan.reduced_darks in (None, {})
    assert scan.reduced_flats in (None, {})
    requires_reduced_dark_and_flat(scan=scan)
    assert len(scan.reduced_darks) == 1
    assert len(scan.reduced_flats) == 1


def test_quick_run_necessary_hdf5(tmpdir):
    """test the `quick_run_necessary` function for NXtomoScan"""
    scan = MockNXtomo(
        scan_path=tmpdir,
        n_proj=20,
        n_ini_proj=20,
        dim=10,
    ).scan
    assert scan.reduced_darks in (None, {})
    assert scan.reduced_flats in (None, {})
    computed = requires_reduced_dark_and_flat(scan=scan)
    assert len(computed) == 2
    assert len(scan.reduced_darks) == 1
    assert len(scan.reduced_flats) == 1



from tomwer.core.volume import EDFVolume, HDF5Volume, RawVolume
from tomwer.core.volume.rawvolume import RawVolumeIdentifier


def test_volume_data_parent_folder():
    edf_volume = EDFVolume(folder="/my/folder/path")
    assert edf_volume.volume_data_parent_folder() == "/my/folder"

    hdf5_volume = HDF5Volume(file_path="/path/to/hdf5/file.hdf5", data_path="entry")
    assert hdf5_volume.volume_data_parent_folder() == "/path/to/hdf5"

    raw_volume = RawVolume(file_path="/path/to/raw.vol")
    assert raw_volume.volume_data_parent_folder() == "/path/to"


def test_raw_identifier():
    raw_volume = RawVolume(file_path="/path/to/raw.vol")
    assert (
        RawVolumeIdentifier.from_str(raw_volume.get_identifier().to_str())
        == raw_volume.get_identifier()
    )



# coding: utf-8

"""
contains gui to select a slice in a volume
"""

from __future__ import annotations


from silx.gui import qt


class ControlWidget(qt.QWidget):
    """
    Widget to lock cor position or compute it or validate it and to
    display the cor value
    """

    sigComputationRequest = qt.Signal()
    """Signal emitted when user request a computation from the settings"""

    sigValidateRequest = qt.Signal()
    """Signal emitted when user validate the current settings"""

    def __init__(self, parent):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QVBoxLayout())

        self._buttons = qt.QWidget(parent=self)
        self._buttons.setLayout(qt.QHBoxLayout())
        self.layout().addWidget(self._buttons)

        spacer = qt.QWidget(self)
        spacer.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Minimum)
        self._buttons.layout().addWidget(spacer)

        self._computeBut = qt.QPushButton("compute", parent=self)
        self._buttons.layout().addWidget(self._computeBut)
        style = qt.QApplication.style()
        applyIcon = style.standardIcon(qt.QStyle.SP_DialogApplyButton)
        self._applyBut = qt.QPushButton(applyIcon, "validate", parent=self)
        self._buttons.layout().addWidget(self._applyBut)
        self.layout().addWidget(self._buttons)

        # make connection
        self._computeBut.pressed.connect(self.sigComputationRequest)
        self._applyBut.pressed.connect(self.sigValidateRequest)



# coding: utf-8
from __future__ import annotations

import os

from tomwer.core.scan.edfscan import EDFTomoScan
from tomwer.core.utils.scanutils import MockEDF


def test_scan_dir(tmpdir):
    full_path = os.path.join(tmpdir, "my", "aquisition", "folder")
    MockEDF.fastMockAcquisition(full_path)
    scan = EDFTomoScan(full_path)
    assert scan.scan_dir_name() == "folder"
    assert scan.scan_basename() == full_path


def test_working_directory():
    """test behavior of the working directory function"""
    scan = EDFTomoScan(scan=None)
    assert scan.working_directory is None
    scan = EDFTomoScan(scan="my_folder")
    assert str(scan.working_directory) == os.path.abspath("my_folder")
    scan = EDFTomoScan(scan="/full/path/to/my/folder")
    assert str(scan.working_directory) == os.path.abspath("/full/path/to/my/folder")



from orangewidget import widget
from tomwer.core.tomwer_object import TomwerObject
from orangewidget.widget import Input, Output


class TomoObjsHubOW(widget.OWBaseWidget, openclass=True):
    name = "Tomo objs hub"
    description = "takes a list of tomo_obj in entry and trigger a tomo_obj scalar for each element in downstream"
    icon = "icons/hub.svg"
    priority = 4000
    keywords = ["hub", "tomo-obj", "tomobj"]

    want_basic_layout = False
    want_control_area = False
    want_main_area = False

    class Inputs:
        tomo_objs = Input("tomo_objs", tuple, default=True, multiple=True)

    class Outputs:
        tomo_objs = Output("tomo_obj", TomwerObject)

    @Inputs.tomo_objs
    def execute(self, tomo_objs, *args, **kwargs):
        if tomo_objs is None:
            return
        for tomo_obj in tomo_objs:
            self.Outputs.tomo_objs.send(tomo_obj)



from __future__ import annotations

import logging
import socket
from contextlib import contextmanager
from multiprocessing import cpu_count
from multiprocessing.pool import ThreadPool

from silx.gui import qt
from tomwer.core.settings import TOMO_BEAMLINES

_logger = logging.getLogger(__name__)


class _PingHosts(qt.QThread):
    def __init__(self, *args, hosts_to_ping: tuple[str], port: int, **kwargs):
        super().__init__(*args, **kwargs)
        self._hosts_to_ping = hosts_to_ping
        self._port = port
        self._accessibility: dict[str, bool] = {}

    @staticmethod
    def can_ping_host(host_name: str, port: int, timeout_s: float = 0.5) -> bool:
        """
        Checks if a host is reachable on a specific TCP port.
        Returns True if reachable, False otherwise.
        """
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.settimeout(timeout_s)
        try:
            sock.connect((host_name, port))
        except (socket.timeout, socket.error):
            return False
        else:
            return True
        finally:
            sock.close()

    def run(self):
        pool = ThreadPool(processes=min(cpu_count(), 4))
        result = pool.map(
            lambda host: self.can_ping_host(host_name=host, port=self._port),
            self._hosts_to_ping,
        )
        pool.close()
        pool.join()
        self._accessibility = {
            host: accessibility
            for host, accessibility in zip(self._hosts_to_ping, result)
        }

    def getHostsAccessibility(self) -> dict[str, bool]:
        return self._accessibility


class RightIconDelegate(qt.QStyledItemDelegate):
    def paint(self, painter, option, index):
        # Set the decorationPosition to Right
        option.decorationPosition = qt.QStyleOptionViewItem.Right
        # Call the base class's paint method to draw the item with the modified option
        super().paint(painter, option, index)


class HostEditor(qt.QComboBox):
    """
    An editable QCombobox that proposes some beamline in order to select an host.

    Use case: blissdata listener
    """

    def __init__(
        self,
        *args,
        hosts: tuple[str] = TOMO_BEAMLINES,
        name: str | None,
        port: int = 25000,
        enable_host_pinging: bool = True,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        if not isinstance(port, int):
            raise TypeError(f"port is expected to be an int. Got {type(port)}")
        self.__hosts = hosts
        self.__port = port
        self.__pingerThread = None
        # thread to check which host are accessible
        self.__enable_host_pinging = enable_host_pinging
        self.addItems(hosts)
        self.setPlaceholderText("id00 or 192.0.2.1")
        # let user use another name or an ip directly
        self.setEditable(True)
        self._updateIcons()

        # set up
        if name:
            self.setCurrentText(name)

        self.__delegate = RightIconDelegate()
        self.setItemDelegate(self.__delegate)

    def _getPingerThread(self):
        return self.__pingerThread

    def setCurrentText(self, *args, **kwargs):
        super().setCurrentText(*args, **kwargs)

    def setCurrentIndex(self, index):
        return super().setCurrentIndex(index)

    def setPort(self, port: int):
        """Defines the port to be used for checking host accessibility"""
        if port != self.__port:
            self.__port = port
            self._updateIcons()

    def __setAccessibleHost(self, host_accessibility: dict[str, bool]):
        """
        Set icons of the ComBox according to the host accessibility.
        Red: cannot reach it, green can.
        """
        with _keepCurrentText(self):
            # for some reason the current text is modified by this section...
            style = qt.QApplication.style()
            accessible_icon = style.standardIcon(qt.QStyle.SP_DialogYesButton)
            not_accessible_icon = style.standardIcon(qt.QStyle.SP_DialogNoButton)
            for host_name, can_be_access in host_accessibility.items():
                idx = self.findText(host_name)
                if idx >= 0:
                    icon = accessible_icon if can_be_access else not_accessible_icon
                    self.setItemIcon(idx, icon)
                else:
                    _logger.error(f"Cannot find host name {host_name}.")
        self.__pingerThread = None

    def _updateIcons(self):
        if not self.__enable_host_pinging:
            return

        if self.__pingerThread is not None:
            _logger.warning("Updating icons is already on-going")
            return

        self.__pingerThread = _PingHosts(
            hosts_to_ping=self.__hosts, port=self.__port, parent=self
        )
        self.__pingerThread.finished.connect(
            lambda: self.__setAccessibleHost(
                self.__pingerThread.getHostsAccessibility()
            )
        )
        self.__pingerThread.start()

    def close(self):
        if self.__pingerThread is not None:
            self.__pingerThread.wait()
        return super().close()


@contextmanager
def _keepCurrentText(w: qt.QComboBox):
    old = w.currentText()
    try:
        yield
    finally:
        w.setCurrentText(old)



import os
import shutil
import tempfile

import numpy

from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt
from tomwer.core.utils.scanutils import MockNXtomo
from tomwer.core.volume.hdf5volume import HDF5Volume
from tomwer.gui.stitching.metadataholder import QStitchingMetadata
from tomwer.gui.stitching.stitching_raw import RawStitchingPlot, AlphaValuesTableWidget


def _createVolumes(axis_0_positions: tuple, root_dir: str):
    volumes = []
    for i_volume, axis_0_position in enumerate(axis_0_positions):
        data = numpy.random.random(20 * 20).reshape(1, 20, 20)
        volume = HDF5Volume(
            file_path=os.path.join(root_dir, f"volume_{i_volume}.hdf5"),
            data_path=f"entry_{i_volume}",
            data=data,
        )
        volume.stitching_metadata = QStitchingMetadata(tomo_obj=volume)
        volume.stitching_metadata.setPxPos(axis=0, value=axis_0_position)
        volumes.append(volume)
    return volumes


def _createScans(axis_0_positions: tuple, root_dir: str):
    scans = []
    for i_scan, axis_0_position in enumerate(axis_0_positions):
        scan = MockNXtomo(
            scan_path=os.path.join(root_dir, f"scan_{i_scan}"),
            n_proj=20,
            n_ini_proj=20,
            dim=10,
        ).scan
        scan.stitching_metadata = QStitchingMetadata(tomo_obj=scan)
        scan.stitching_metadata.setPxPos(axis=0, value=axis_0_position)
        scans.append(scan)
    return scans


class TestPlotRawStitching(TestCaseQt):
    """Test RawStitchingPlot widget"""

    def setUp(self):
        super().setUp()
        self._tmp_path = tempfile.mkdtemp()
        self.widget = RawStitchingPlot(alpha_values=True)

    def tearDown(self):
        self.widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.widget.close()
        self.widget = None
        shutil.rmtree(self._tmp_path)

    def testWithVolumes(self):
        axis_0_positions = (0, 12, 50)
        volumes = _createVolumes(
            axis_0_positions=axis_0_positions, root_dir=self._tmp_path
        )
        assert len(volumes) == 3
        self.widget.addTomoObj(volumes[0])
        images = self.widget.getAllImages()
        assert len(images) == 1

        self.widget.setTomoObjs(volumes)
        images = self.widget.getAllImages()
        assert len(images) == len(volumes)
        # silx is storing origins as x, y
        origins = tuple([image.getOrigin()[1] for image in images])
        assert tuple(sorted(origins)) == axis_0_positions

    def testWithScans(self):
        axis_0_positions = (0, 10, 20, 34)
        scans = _createScans(axis_0_positions=axis_0_positions, root_dir=self._tmp_path)
        self.widget.setActive(False)
        self.widget.setTomoObjs(scans)
        assert len(self.widget.getAllImages()) == 0
        self.widget.setActive(True)
        self.qapp.processEvents()
        assert len(self.widget.getAllImages()) == len(scans)


class TestAlphaValuesTableWidget(TestCaseQt):
    """test AlphaValuesTableWidget"""

    def setUp(self):
        super().setUp()
        self._tmp_path = tempfile.mkdtemp()
        self.widget = AlphaValuesTableWidget()

    def tearDown(self):
        self.widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.widget.close()
        self.widget = None
        shutil.rmtree(self._tmp_path)

    def testSettingTomoObj(self):
        self.widget.setTomoObjs(
            _createVolumes(axis_0_positions=(0, 12, 50), root_dir=self._tmp_path),
        )

        assert len(self.widget._sliders) == 3
        self.widget.setTomoObjs(
            _createScans(axis_0_positions=(89, 78), root_dir=self._tmp_path)
        )
        assert len(self.widget._sliders) == 2



import os
import shutil
import numpy

from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.core.utils.scanutils import MockNXtomo
from tomwer.tests.datasets import TomwerCIDatasets
from tomwer.core.volume.hdf5volume import HDF5Volume
from tomwer.core.volume.edfvolume import EDFVolume
from tomwer.core.process.reconstruction.output import (
    PROCESS_FOLDER_RECONSTRUCTED_SLICES,
)
from nxtomomill.converter.hdf5.utils import PROCESSED_DATA_DIR_NAME, RAW_DATA_DIR_NAME


def test_scan_dir(tmp_path):
    """test result of NxtomoScan.scan_dir_name is correct"""
    tmp_dir = tmp_path / "frm_edftomomill_twoentries_dataset"
    tmp_dir.mkdir()
    tmp_dir = str(tmp_dir)
    dataset_file = os.path.join(tmp_dir, "frm_edftomomill_twoentries.nx")
    shutil.copyfile(
        TomwerCIDatasets.get_dataset(
            "h5_datasets/frm_edftomomill_twoentries.nx",
        ),
        dataset_file,
    )
    assert os.path.isfile(dataset_file)
    scan = NXtomoScan(scan=dataset_file, entry="entry0000")
    assert scan.scan_dir_name() == tmp_dir.split(os.sep)[-1]


def test_flat_field_interval(tmp_path):
    """test the call to ff_interval"""
    tmp_dir = tmp_path / "test_flat_field_interval"
    scan_path = os.path.join(tmp_dir, "my_scan_1")
    scan_1 = MockNXtomo(
        scan_path=scan_path,
        n_ini_proj=20,
        n_proj=20,
        n_alignement_proj=2,
        create_final_flat=True,
        create_ini_dark=True,
        create_ini_flat=True,
        n_refs=5,
    ).scan
    numpy.testing.assert_equal(scan_1.ff_interval, 20)

    scan_path2 = os.path.join(tmp_dir, "my_scan_2")
    scan_2 = MockNXtomo(
        scan_path=scan_path2,
        n_ini_proj=10,
        n_proj=10,
        n_alignement_proj=2,
        create_final_flat=False,
        create_ini_dark=True,
        create_ini_flat=True,
        n_refs=1,
    ).scan
    numpy.testing.assert_equal(scan_2.ff_interval, 0)


def test_working_directory():
    """test behavior of the working directory function"""
    scan = NXtomoScan(scan=None, entry="my_entry")
    assert scan.working_directory is None
    scan = NXtomoScan(scan="/full/path/my_file.sh", entry="my_entry")
    assert str(scan.working_directory) == os.path.abspath("/full/path")
    scan = NXtomoScan(scan="my_file.sh", entry="my_entry")
    assert str(scan.working_directory) == os.path.abspath(".")


def test_get_reconstructions_slices(tmp_path):
    """test NXtomoscan.get_reconstruction_urls and make sure all related scan can be found from a scan
    (as long as they are stored at a 'searched' location)
    """
    test_dir = tmp_path / "test_get_reconstructions_urls"
    test_dir.mkdir()
    raw_data_dir = test_dir / RAW_DATA_DIR_NAME / "frm_edftomomill_twoentries"
    raw_data_dir.mkdir(parents=True)
    processed_data_dir = (
        test_dir / PROCESSED_DATA_DIR_NAME / "frm_edftomomill_twoentries"
    )
    processed_data_dir.mkdir(parents=True)
    third_path_dir = test_dir / "third_part"
    third_path_dir.mkdir(parents=True)

    output_file_path = os.path.join(raw_data_dir, "frm_edftomomill_twoentries.nx")
    shutil.copyfile(
        TomwerCIDatasets.get_dataset(
            "h5_datasets/frm_edftomomill_twoentries.nx",
        ),
        output_file_path,
    )
    scan = NXtomoScan(
        scan=output_file_path,
        entry="entry0000",
    )
    assert str(scan.working_directory) == str(raw_data_dir)
    assert os.path.exists(output_file_path)

    # create a edf volume in PROCESSED_DATA
    data = numpy.linspace(start=0, stop=100, num=100).reshape(  # pylint: disable=E1121
        1, 10, 10
    )
    edf_volume = EDFVolume(
        folder=os.path.join(
            raw_data_dir,
            PROCESS_FOLDER_RECONSTRUCTED_SLICES,
        ),
        volume_basename="frm_edftomomill_twoentries_slice",
        data=data,
    )
    edf_volume.save()

    # create an hdf5 volume in RAW_DATA
    hdf5_volume = HDF5Volume(
        file_path=os.path.join(
            processed_data_dir,
            PROCESS_FOLDER_RECONSTRUCTED_SLICES,
            "frm_edftomomill_twoentries_slice.h5",
        ),
        data_path="/entry0000",
        data=data,
    )
    hdf5_volume.save()

    # create a third hdf5 volume in 'third_part_dir'
    third_part_volume = HDF5Volume(
        file_path=os.path.join(
            third_path_dir,
            PROCESS_FOLDER_RECONSTRUCTED_SLICES,
            "frm_edftomomill_twoentries_slice.h5",
        ),
        data_path="entry0000",
        data=data,
    )
    third_part_volume.save()

    reconstructions_urls = scan.get_reconstructed_slices()
    assert len(reconstructions_urls) == 2
    assert edf_volume.get_identifier().to_str() in reconstructions_urls
    assert hdf5_volume.get_identifier().to_str() in reconstructions_urls
